# Generated by Django 4.2.16 on 2024-11-26 01:48

import ansible_base.resource_registry.models.service_identifier
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('dab_resource_registry', '0006_alter_resource_service_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='resource',
            name='ansible_id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, help_text='A unique ID identifying this resource by the resource server.', unique=True),
        ),
        migrations.AlterField(
            model_name='resource',
            name='content_type',
            field=models.ForeignKey(help_text='The content type for this resource.', on_delete=django.db.models.deletion.CASCADE, related_name='resources', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='resource',
            name='is_partially_migrated',
            field=models.BooleanField(default=False, help_text="A flag indicating that the resource has been copied into the resource server, but the service_id hasn't been updated yet."),
        ),
        migrations.AlterField(
            model_name='resource',
            name='name',
            field=models.CharField(help_text='The name of this resource.', max_length=512, null=True),
        ),
        migrations.AlterField(
            model_name='resource',
            name='object_id',
            field=models.TextField(db_index=True, help_text='The object id for this resource.'),
        ),
        migrations.AlterField(
            model_name='resource',
            name='service_id',
            field=models.UUIDField(default=ansible_base.resource_registry.models.service_identifier.service_id, help_text='ID of the service responsible for managing this resource.'),
        ),
        migrations.AlterField(
            model_name='resourcetype',
            name='content_type',
            field=models.OneToOneField(help_text='The content type for this resource type.', on_delete=django.db.models.deletion.CASCADE, related_name='resource_type', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='resourcetype',
            name='externally_managed',
            field=models.BooleanField(help_text='Is this resource type managed externally from this service.'),
        ),
        migrations.AlterField(
            model_name='resourcetype',
            name='name',
            field=models.CharField(db_index=True, editable=False, help_text='The name of this resource type.', max_length=256, unique=True),
        ),
    ]



class ResourceTypeProcessor:
    """
    This class allows services to customize how objects are serialized and
    saved in the resources API in cases where the underlying model is not
    exactly the same.

    A good example of this is team and organization memberships. Galaxy and
    EDA implement membership on an organization as a many to many field,
    whereas AWX uses roles to track membership.
    """

    def __init__(self, instance):
        self.instance = instance

    def pre_serialize(self):
        """
        This gets called on an instance of a model before it is sent in the
        `instance` kwarg to the ResourceType serializer. This can be customized
        to add additional fields onto the model object that are expect in the
        resource type serializer, but aren't on the local version of the model.
        """
        return self.instance

    def pre_serialize_additional(self):
        """
        Same as pre_serialize, but is called before ADDITIONAL_DATA_SERIALIZER
        is instantiated.
        """
        raise NotImplementedError("Additional data is not supported by default.")

    def save(self, validated_data, is_new=False):
        """
        This gets called when an instance of a Resource is saved and allows for
        services to customize how the resource gets saved with their local copy
        of the model.
        """
        for k, val in validated_data.items():
            setattr(self.instance, k, val)

        self.instance.save()
        return self.instance



import logging
import threading
from contextlib import contextmanager

logger = logging.getLogger('ansible_base.activitystream.signals')


class ActivityStreamEnabled(threading.local):
    def __init__(self):
        self.enabled = True

    def __bool__(self):
        return self.enabled


activitystream_enabled = ActivityStreamEnabled()


@contextmanager
def no_activity_stream():
    previous_value = activitystream_enabled.enabled
    activitystream_enabled.enabled = False
    try:
        yield
    finally:
        activitystream_enabled.enabled = previous_value


def _store_activitystream_entry(old, new, operation):
    if not activitystream_enabled:
        return

    from ansible_base.activitystream.models import Entry
    from ansible_base.lib.utils.models import diff

    if operation not in ('create', 'update', 'delete'):
        raise ValueError("Invalid operation: {}".format(operation))

    excluded = getattr(new, 'activity_stream_excluded_field_names', [])
    limit = getattr(new, 'activity_stream_limit_field_names', [])
    delta = diff(old, new, exclude_fields=excluded, limit_fields=limit, all_values_as_strings=True)

    if not delta:
        # No changes to store
        return

    # If only one of old or new is None, then use the existing one as content_object
    # The case where both are None is handled above (no changes to store)
    if new is None:
        content_object = old
    else:
        content_object = new

    return Entry.objects.create(
        content_object=content_object,
        operation=operation,
        changes=delta.dict(),
    )


def _store_activitystream_m2m(given_instance, model, operation, pk_set, reverse, field_name):
    if not activitystream_enabled:
        return

    from ansible_base.activitystream.models import Entry
    from ansible_base.lib.utils.models import current_user_or_system_user

    if operation not in ('associate', 'disassociate'):
        raise ValueError("Invalid operation: {}".format(operation))

    instances = model.objects.filter(pk__in=pk_set)
    user = current_user_or_system_user()
    entries = []

    for instance in instances:
        entry = Entry(
            content_object=instance if reverse else given_instance,
            operation=operation,
            related_content_object=given_instance if reverse else instance,
            related_field_name=field_name,
            created_by=user,
        )
        entries.append(entry)

    Entry.objects.bulk_create(entries)


# post_save
def activitystream_create(sender, instance, created, **kwargs):
    """
    This signal is registered via the activity stream AuditableModel abstract
    model/class. It is called after save() of any model that inherits from
    AuditableModel. (It is registered as a post_save signal.)

    This signal only handles creation of new objects (created=True). For
    updates, use the activitystream_update signal, where we can compare the
    old and new objects to determine what has changed.
    """
    if not created:
        # We only want to create an activity stream entry for new objects
        # Update events are handled by the activitystream_update receiver
        return

    _store_activitystream_entry(None, instance, 'create')


# pre_save
def activitystream_update(sender, instance, raw, using, update_fields, **kwargs):
    """
    This signal is registered via the activity stream AuditableModel abstract
    model/class. It is called before save() of any model that inherits from
    AuditableModel. (It is registered as a pre_save signal.)

    This signal only handles updates of existing objects. For creation of
    objects, see the above activitystream_create().
    """
    if instance.pk is None:
        # We only want to create an activity stream entry for existing objects
        # Creation events are handled by the activitystream_create receiver
        return

    try:
        old = sender.objects.get(pk=instance.pk)
    except sender.DoesNotExist:
        return

    _store_activitystream_entry(old, instance, 'update')


# pre_delete
def activitystream_delete(sender, instance, using, origin, **kwargs):
    """
    This signal is registered via the activity stream AuditableModel abstract
    model/class. It is called before delete() of any model that inherits from
    AuditableModel. (It is registered as a pre_delete signal.)
    """
    if instance.pk is None:
        return

    _store_activitystream_entry(instance, None, 'delete')


# m2m_changed
def activitystream_m2m_changed(sender, instance, action, reverse, model, pk_set, **kwargs):
    """
    This signal is registered via the activity stream AuditableModel abstract
    model/class. It is called when a many-to-many relationship is changed
    (added or removed) for any model that inherits from AuditableModel. (It is
    registered as a m2m_changed signal.)
    """
    if action not in ('post_add', 'post_remove', 'pre_clear'):
        return

    if 'field_name' not in kwargs:
        # Theory says we should never get here, the field name is established when the signal is connected.
        raise ValueError(
            f"Missing field_name in kwargs while trying to store activity stream {action} event for instance={instance}, model={model}, sender={sender}"
        )

    field_name = kwargs['field_name']
    operation = 'associate' if action == 'post_add' else 'disassociate'

    if action == 'pre_clear':
        # This is called if someone calls .clear() on a m2m field. But we need to handle the forward and reverse
        # relations differently.
        if reverse:
            # Okay. We need to talk. Just you - the reader trying to understand this code - and I.
            # Look. We want to always store the activity stream entry on the forward relation.
            # Let's assume we have an Animal model with a 'people_friends' field which is a m2m pointing to User.
            # This is the forward relation.
            # If we do: user.animal_friends.clear() - the reverse relation - we need to get the PKs of
            # every animal that is being removed from the user's animal_friends.
            # Note that in this case, model is the Animal model, and instance is the user.
            pk_set = model.objects.filter(**{field_name: instance}).values_list('pk', flat=True)
        else:
            # If we're not reversing, then we're clearing the forward relation. So it's easy to get the PKs,
            # given we have the field name and the instance.
            pk_set = getattr(instance, field_name).all().values_list('pk', flat=True)

    _store_activitystream_m2m(instance, model, operation, pk_set, reverse, field_name)



"""Organization models."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .common import UniqueNamedCommonModel


class AbstractOrganization(UniqueNamedCommonModel):
    """An abstract base class for organizations."""

    class Meta:
        abstract = True

    description = models.TextField(
        null=False,
        default="",
        blank=True,
        help_text=_("The organization description."),
    )



from datetime import datetime, timedelta

import jwt

from ansible_base.resource_registry.models import Resource, service_id
from ansible_base.resource_registry.resource_server import get_resource_server_config
from ansible_base.resource_registry.utils.sso_provider import get_sso_provider_server


def get_user_auth_code(user, social_user=None, oidc_alt_key=None):
    """
    Generate an authentication code using the service's configured secret key.

    user: Django User instance
    social_user: SocialUser or AuthenticatorUser instance for the backend used to authenticate the user.
    oidc_alt_key: for some OIDC backends, the sub claim on the JWT gets used as the UID rather than the username.
        This param lets us pass the sub id to the auth_code, which can be used as a secondary method for
        account lookups.
    """
    config = get_resource_server_config()
    payload = {
        "iss": str(service_id()),
        "type": "auth_code",
        "username": user.username,
        "sub": str(Resource.get_resource_for_object(user).ansible_id),
        "exp": datetime.now() + timedelta(seconds=15),
        "sso_uid": None,
        "sso_backend": None,
        "sso_server": None,
        "oidc_alt_key": oidc_alt_key,
        "auth_backend": getattr(user, "backend", None),
    }

    if social_user is not None:
        if isinstance(social_user.provider, str):
            provider = social_user.provider
        else:
            provider = social_user.provider.slug

        server, uid = get_sso_provider_server(provider, social_user.uid)

        payload["sso_uid"] = uid
        payload["sso_backend"] = provider
        payload["sso_server"] = server

    return jwt.encode(payload, config["SECRET_KEY"], config["JWT_ALGORITHM"])



from rest_framework.viewsets import ModelViewSet

from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView
from ansible_base.lib.utils.views.permissions import IsSuperuserOrAuditor
from ansible_base.oauth2_provider.models import OAuth2Application
from ansible_base.oauth2_provider.permissions import OAuth2ScopePermission
from ansible_base.oauth2_provider.serializers import OAuth2ApplicationSerializer


class OAuth2ApplicationViewSet(AnsibleBaseDjangoAppApiView, ModelViewSet):
    queryset = OAuth2Application.objects.all()
    serializer_class = OAuth2ApplicationSerializer
    permission_classes = [OAuth2ScopePermission, IsSuperuserOrAuditor]



# Generated by Django 4.2.11 on 2024-06-06 17:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0011_authenticatormap_role_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticatormap',
            name='map_type',
            field=models.CharField(choices=[('allow', 'allow'), ('is_superuser', 'is_superuser'), ('role', 'role'), ('organization', 'organization'), ('team', 'team')], default='team', help_text='What will the map grant the user? System access (allow) a team or organization membership, the superuser flag or a role in the system', max_length=17),
        ),
    ]



import oauth2_provider.models as oauth2_models
from oauth2_provider.contrib.rest_framework import TokenHasReadWriteScope
from rest_framework.permissions import BasePermission, IsAuthenticated

# This is not in views/permissions.py because we import that in other oauth2_provider files
# and when we try to specify this OAuth2ScopePermission in an app's settings.py, we get a
# cyclic dependency.


class OAuth2ScopePermission(BasePermission):
    """
    A DRF Permission class to be used by apps that functions in the following way:

    - If an OAuth 2 token is used to authenticate, then its scope must contain the required scope
      (i.e. "read" cannot use "unsafe" methods)
    - Otherwise, fall back.
    """

    def has_permission(self, request, view):
        is_authenticated = IsAuthenticated().has_permission(request, view)
        is_oauth = False
        has_oauth_permission = False
        if is_authenticated and request.auth and isinstance(request.auth, oauth2_models.AbstractAccessToken):
            is_oauth = True
            scopes = request.auth.scope.split()
            if 'write' in scopes and 'read' not in scopes:
                request.auth.scope += ' read'  # write implies read
            token_permission = TokenHasReadWriteScope()
            has_oauth_permission = token_permission.has_permission(request, view)
        return is_authenticated and (not is_oauth or has_oauth_permission)



import copy
import os
import uuid
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime, timedelta
from unittest import mock

import pytest
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from rest_framework.test import APIClient

from ansible_base.lib.testing.util import copy_fixture, delete_authenticator


@pytest.fixture
def randname():
    def _randname(prefix):
        return f"{prefix}-{uuid.uuid4().hex[:6]}"

    return _randname


@pytest.fixture
def env():
    """
    Set an environment variable within a context manager.
    """

    @contextmanager
    def _env(key, value):
        old_value = os.environ.get(key)
        os.environ[key] = value
        try:
            yield
        finally:
            if old_value is None:
                del os.environ[key]
            else:
                os.environ[key] = old_value

    return _env


@pytest.fixture
def settings_override_mutable(settings):
    """
    pytest-django's settings  doesn't handle mutable settings types well
    (https://github.com/pytest-dev/pytest-django/issues/601). This fixture is an
    attempt to solve that problem.

    Use as follows:

    def test_foo(settings, settings_override_mutable):
        with settings_override_mutable("SOME_SETTING_VAR"):
            settings.SOME_SETTING_VAR['and']['nested']['a']['bit'] = "new_value"
            # or even
            delattr(settings, 'SOME_SETTING_VAR')  # for some reason normal "del" doesn't work here
        # But out of the context manager the whole var is restored
        assert settings.SOME_SETTING_VAR['and']['nested']['a']['bit'] == "old_value"
    """

    @contextmanager
    def f(setting_key):
        original = copy.deepcopy(getattr(settings, setting_key))
        try:
            yield
        finally:
            setattr(settings, setting_key, original)

    return f


@pytest.fixture
def local_authenticator(db):
    from ansible_base.authentication.models import Authenticator

    authenticator = Authenticator.objects.create(
        name="Test Local Authenticator",
        enabled=True,
        create_objects=True,
        remove_users=False,
        type="ansible_base.authentication.authenticator_plugins.local",
        configuration={},
    )
    return authenticator


@pytest.fixture
def unauthenticated_api_client(db):
    return APIClient()


@pytest.fixture
def admin_api_client(db, admin_user, local_authenticator):
    # We don't use the is_staff flag anywhere. Instead we use is_superuser. This can
    # cause some permission checks to unexpectedly break in production where this flag
    # never gets set to true.
    admin_user.is_staff = False
    admin_user.save()
    client = APIClient()
    client.login(username="admin", password="password")
    yield client
    try:
        client.logout()
    except AttributeError:
        # The test might have logged the user out already (e.g. to test the logout signal)
        pass


@pytest.fixture
def user(db, django_user_model, local_authenticator):
    return django_user_model.objects.create_user(username="user", password="password")


@copy_fixture(copies=3)
@pytest.fixture
def random_user(db, django_user_model, randname, local_authenticator):
    return django_user_model.objects.create_user(username=randname("user"), password="password")


@pytest.fixture
def user_api_client(db, user, local_authenticator):
    client = APIClient()
    client.login(username="user", password="password")
    yield client
    try:
        client.logout()
    except AttributeError:
        # The test might have logged the user out already (e.g. to test the logout signal)
        pass


@pytest.fixture
def no_log_messages():
    """
    This fixture returns a function (a context manager) which allows you to disable
    logging for a very specific part of a test.
    """

    @contextmanager
    def f():
        import logging

        logging.disable(logging.CRITICAL)
        try:
            yield
        finally:
            logging.disable(logging.NOTSET)

    return f


@pytest.fixture
def shut_up_logging(no_log_messages):
    """
    This fixture allows you to temporarily disable logging for an entire test.
    """
    with no_log_messages():
        yield


@pytest.fixture
def expected_log(no_log_messages):
    """
    This fixture returns a function (a context manager) which allows you to assert
    that a logger is called appropriately for a line or block of code.

    Use it as a fixture, and then in your test:

    with expected_log("path.to.logger", "info", "some substring"):
        # code that should trigger the log message

    Or you can use functools.partial to make it even more concise if you're
    testing the same logger a bunch of times:

    expected_log = partial(expected_log, "path.to.logger")

    with expected_log("info", "some substring"):
        # code that should trigger the log message
    """

    @contextmanager
    def f(patch, severity, substr, assert_not_called=False):
        with mock.patch(f'{patch}.{severity}') as logger:
            with no_log_messages():
                yield

            call_count = sum(1 for call in logger.call_args_list if substr in call.args[0])

            if assert_not_called:
                assert call_count == 0, f"Expected 0 calls but got {call_count}"
            else:
                assert call_count == 1, f"Expected 1 call but got {call_count}"

    return f


@pytest.fixture
def rsa_keypair_factory():
    """
    This fixture returns a factory function that creates a new RSA keypair every time it is called.

    This is useful for tests that need to create multiple RSA keypairs. @copy_fixture isn't suitable
    because it only ends up creating one keypair, so this avoids the issue.

    A default instance of this fixture is available as rsa_keypair.
    """

    def rsa_keypair():
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        public_key = private_key.public_key()
        private_key_bytes = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode("utf-8")
        public_key_bytes = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ).decode("utf-8")
        RSAKeyPair = namedtuple("RSAKeyPair", ["private", "public"])
        return RSAKeyPair(private=private_key_bytes, public=public_key_bytes)

    return rsa_keypair


@pytest.fixture
def rsa_keypair(rsa_keypair_factory):
    return rsa_keypair_factory()


@copy_fixture(copies=3)
@pytest.fixture
def rsa_keypair_with_signed_cert(rsa_keypair_factory):
    root_keypair = rsa_keypair_factory()
    root_private_key = serialization.load_pem_private_key(root_keypair.private.encode("utf-8"), password=None)
    root_public_key = serialization.load_pem_public_key(root_keypair.public.encode("utf-8"))
    rsa_keypair = rsa_keypair_factory()
    public_key = serialization.load_pem_public_key(rsa_keypair.public.encode("utf-8"))
    issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
            x509.NameAttribute(NameOID.COMMON_NAME, u"mycompany.com"),
        ]
    )
    subject = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
            x509.NameAttribute(NameOID.COMMON_NAME, u"qa.mycompany.com"),
        ]
    )
    root_certificate = (
        x509.CertificateBuilder()
        .subject_name(issuer)
        .issuer_name(issuer)
        .public_key(root_public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow())
        .not_valid_after(datetime.utcnow() + timedelta(days=365))
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(u"mycompany.com")]),
            critical=False,
        )
        .sign(root_private_key, hashes.SHA256())
    )
    certificate = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow())
        .not_valid_after(datetime.utcnow() + timedelta(days=365))
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(u"qa.mycompany.com")]),
            critical=False,
        )
        .sign(root_private_key, hashes.SHA256())
    )

    root_certificate_bytes = root_certificate.public_bytes(serialization.Encoding.PEM).decode("utf-8")
    certificate_bytes = certificate.public_bytes(serialization.Encoding.PEM).decode("utf-8")

    RSAKeyPairWithCert = namedtuple("RSAKeyPairWithCert", ["private", "public", "certificate"])
    CertificateChain = namedtuple("CertificateChain", ["root", "subordinate"])

    root_keypair_with_cert = RSAKeyPairWithCert(private=root_keypair.private, public=root_keypair.public, certificate=root_certificate_bytes)
    subordinate_keypair_with_cert = RSAKeyPairWithCert(private=rsa_keypair.private, public=rsa_keypair.public, certificate=certificate_bytes)
    return CertificateChain(root=root_keypair_with_cert, subordinate=subordinate_keypair_with_cert)


@copy_fixture(copies=3)
@pytest.fixture
def rsa_keypair_with_cert(rsa_keypair_factory):
    rsa_keypair = rsa_keypair_factory()
    private_key = serialization.load_pem_private_key(rsa_keypair.private.encode("utf-8"), password=None)
    public_key = serialization.load_pem_public_key(rsa_keypair.public.encode("utf-8"))
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
            x509.NameAttribute(NameOID.COMMON_NAME, u"mycompany.com"),
        ]
    )
    certificate = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow())
        .not_valid_after(datetime.utcnow() + timedelta(days=365))
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(u"mycompany.com")]),
            critical=False,
        )
        .sign(private_key, hashes.SHA256())
    )
    certificate_bytes = certificate.public_bytes(serialization.Encoding.PEM).decode("utf-8")
    RSAKeyPairWithCert = namedtuple("RSAKeyPairWithCert", ["private", "public", "certificate"])
    return RSAKeyPairWithCert(private=rsa_keypair.private, public=rsa_keypair.public, certificate=certificate_bytes)


@pytest.fixture
def ldap_configuration():
    return {
        "SERVER_URI": ["ldap://ldap06.example.com:389"],
        "BIND_DN": "cn=ldapadmin,dc=example,dc=org",
        "BIND_PASSWORD": "securepassword",
        "START_TLS": False,
        "CONNECTION_OPTIONS": {"OPT_REFERRALS": 0, "OPT_NETWORK_TIMEOUT": 30},
        "USER_SEARCH": ["ou=users,dc=example,dc=org", "SCOPE_SUBTREE", "(cn=%(user)s)"],
        "USER_DN_TEMPLATE": "cn=%(user)s,ou=users,dc=example,dc=org",
        "USER_ATTR_MAP": {"email": "mail", "last_name": "sn", "first_name": "givenName"},
        "GROUP_SEARCH": ["ou=groups,dc=example,dc=org", "SCOPE_SUBTREE", "(objectClass=groupOfNames)"],
        "GROUP_TYPE": "MemberDNGroupType",
        "GROUP_TYPE_PARAMS": {"name_attr": "cn", "member_attr": "member"},
    }


@pytest.fixture
def ldap_authenticator(ldap_configuration):
    from ansible_base.authentication.models import Authenticator

    authenticator = Authenticator.objects.create(
        name="Test LDAP Authenticator",
        enabled=True,
        create_objects=True,
        remove_users=True,
        type="ansible_base.authentication.authenticator_plugins.ldap",
        configuration=ldap_configuration,
    )
    yield authenticator
    delete_authenticator(authenticator)


@pytest.fixture
def create_mock_method():
    # Creates a function that when called, generates a function that can be used to patch
    # a method. Useful for when you want to mock an object method that modifies an object's state.
    # Accepts arguments:
    #   field_dicts: iterable of dictionaries containing key-value pairs to modify for parent object
    def mock_method_generator(field_dicts: list[dict]):
        def create_generator():
            for field_dict in field_dicts:
                yield field_dict

        generator = create_generator()

        def mock_method(self, ignore_cache=False):
            # Get dictionary of fields to modify
            fields = next(generator)
            # Modify JWTCert object with each field defined in the fields dictionary
            for field, value in fields.items():
                setattr(self, field, value)

        return mock_method

    return mock_method_generator



from typing import Optional

from django.utils.encoding import smart_str


def make_json_safe(value):
    if isinstance(value, (list, dict, str, int, float, bool, type(None))):
        return value

    return smart_str(value)


def is_empty(value: Optional[str]) -> bool:
    """Checks if the value is an empty string (stripping whitespaces)"""
    return value is None or str(value).strip() == ''



from django.utils.translation import gettext

'''
Simple helper class that can be used to translate messages or not translate messages.
Ex: Function that calls a raise ValidationError and logger.error one after the other.
'''


class translatableConditionally:
    def __init__(self, message):
        self.message = message

    def not_translated(self):
        return self.message

    def translated(self):
        return gettext(self.message)



from django.urls import include, path

from ansible_base.rbac.api.router import router
from ansible_base.rbac.api.views import RoleMetadataView
from ansible_base.rbac.apps import AnsibleRBACConfig

app_name = AnsibleRBACConfig.label

api_version_urls = [
    path('', include(router.urls)),
    path(r'role_metadata/', RoleMetadataView.as_view(), name="role-metadata"),
]

root_urls = []

api_urls = []



import base64
import hashlib
import json
import logging
import re
from typing import Any, Optional, Tuple

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from django.conf import settings
from django.utils.encoding import smart_bytes, smart_str
from django.utils.functional import SimpleLazyObject

__all__ = [
    'ENCRYPTED_STRING',
    'ansible_encryption',
]

logger = logging.getLogger('ansible_base.lib.utils.encryption')


ENCRYPTED_STRING = '$encrypted$'
ENCRYPTION_METHOD = 'AESCBC'

regex = '^(?:' + re.escape(ENCRYPTED_STRING) + ')(?:UTF8\\$)?([^(UTF8)]+)\\$((?:[A-Za-z0-9+/]{4})+(?:[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?)$'
"""
Explaining this regex, you're welcome :)

Our encrypted strings contain the following segments:
    $encrypted$ - our encrypted marker in the variable ENCRYPTED_STRING
    UTF8$ - an Optional UTF-8 marker
    AESCBC$ - The encryption algorithm specified by ENCRYPTION_METHOD
    <base64 encoded data>

Note base64 encoded data is "tuples" of 4 characters where characters are defined as [a-zA-Z0-9+/]; with the last tuple either being:
  * 4 characters
  * 3 characters + '='
  * 2 characters + '=='

In a regex we can group things together with () this will cause a matched group.
If we don't need to retrieve the matched group we can use (?:...)

Here is the breakdown of the regex:
  * ^ - The beginning of the string
  * (?:'+ re.escape(ENCRYPTED_STRING) +') - our ENCRYPTED_STRING value (we use re.escape on it because it has $ in it), don't retrieve group
  * (?:UTF8\\$)? - an optional UTF8$ string, don't retrieve group
  * ([^(UTF8)]+)\\$ - One or more group of characters that are not 'UTF8' followed by $, retrieve the group (this is the algorithm)
  * ( - open a group that will contain all of the base64 encoded data
      * (?:[A-Za-z0-9+/]{4})+ - One or more group of 4 base64 characters, don't retrieve group
      * (?:[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)? - 0 or 1 ending tuples in the following formats, don't retrieve group:
          * 3 base64 characters + '='
          * 2 base64 characters + '=='
  * ) - the close of the retrieved base64 group
  * $ - the end of the string

credit: the mastermind behind this regex is John Westcott aka CodeCaptainAwesome. Thank you John :)
"""
ENCRYPTED_REGEX = re.compile(regex)


class Fernet256(Fernet):
    """Not technically Fernet, but uses the base of the Fernet spec and uses AES-256-CBC
    instead of AES-128-CBC. All other functionality remain identical.
    """

    def __init__(self):
        h = hashlib.sha512()
        h.update(smart_bytes(settings.SECRET_KEY))
        self.key = h.digest()

        if len(self.key) != 64:
            raise ValueError("Fernet key must be 64 url-safe base64-encoded bytes.")

        self._signing_key = self.key[:32]
        self._encryption_key = self.key[32:]
        self._backend = default_backend()

    def is_encrypted_string(self, value: Any) -> Tuple[bool, Optional[str], Optional[str]]:
        """
        Performs a partial test to see if the string is encrypted by our algorithm
        Note: this can fail because some strings resemble base 64 encoded ones
            for instance, if user's secret is $encrypted$UTF8$AESCBC$junk, the code will break
        Returns the Tuple(is_encrypted, encryption_algorithm, data)
        """
        if not isinstance(value, str):
            value = str(value)

        # Use the precompiled regex to check for encrypted string format
        if encrypted_match := ENCRYPTED_REGEX.match(value):
            return True, encrypted_match.group(1), encrypted_match.group(2)
        else:
            return False, None, None

    def encrypt_string(self, value: Any) -> str:
        # Its possible for a serializer to accept a number for a CharField (like 5). In the serializer its "5" but when we get here it might be 5

        is_encrypted = self.is_encrypted_string(value)[0]

        if is_encrypted:
            return value
        value = json.dumps(value)

        encrypted = self.encrypt(smart_bytes(value))
        b64data = smart_str(base64.b64encode(encrypted))
        return f'{ENCRYPTED_STRING}UTF8${ENCRYPTION_METHOD}${b64data}'

    def decrypt_string(self, value: str) -> Any:
        if not isinstance(value, str):
            raise ValueError("decrypt_string can only accept string")

        is_encrypted, algo, data = self.is_encrypted_string(value)

        if not is_encrypted:
            return value

        # Ensure the algorithm is what we expect
        if algo != ENCRYPTION_METHOD:
            raise ValueError(f'Unsupported algorithm: {algo}')

        # Finally decode the value
        encrypted = base64.b64decode(data)

        decrypted_value = smart_str(self.decrypt(encrypted))
        try:
            return json.loads(decrypted_value)
        except json.JSONDecodeError as e:
            logger.exception("Failed to decode encrytped value as json from database")
            raise e


ansible_encryption = SimpleLazyObject(func=lambda: Fernet256())



import logging
from typing import Optional, Type

from django.conf import settings
from django.db.models import Model
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

logger = logging.getLogger('ansible_base.rbac.managed')


class ManagedRoleConstructor:
    """Subclasses must define attributes, or override methods that use attribues
    - name
    - description
    - model_name
    - permission_list
    """

    def __init__(self, overrides=None):
        if overrides:
            for key, value in overrides.items():
                setattr(self, key, value)

    def get_model(self, apps):
        "It is intended that this will error if subclass did not set model_name"
        if self.model_name is None:
            return None
        return apps.get_model(self.model_name)

    def get_permissions(self, apps) -> set[str]:
        "It is intended that this will error if subclass did not set permission_list"
        return self.permission_list

    def get_translated_name(self) -> str:
        return _(self.name)

    def get_content_type(self, apps):
        model = self.get_model(apps)
        if model is None:
            return None
        content_type_cls = apps.get_model('contenttypes', 'ContentType')
        return content_type_cls.objects.get_for_model(model)

    def get_or_create(self, apps):
        "Create from a list of text-type permissions and do validation"
        role_definition_cls = apps.get_model('dab_rbac', 'RoleDefinition')
        defaults = {
            'description': self.description,
            'content_type': self.get_content_type(apps),
            'managed': True,
        }
        rd, created = role_definition_cls.objects.get_or_create(name=self.name, defaults=defaults)

        if created:
            permissions = self.get_permissions(apps)
            permission_cls = apps.get_model('dab_rbac', 'DABPermission')
            perm_list = [permission_cls.objects.get(codename=str_perm) for str_perm in permissions]
            rd.permissions.add(*perm_list)
            logger.info(f'Created {self.shortname} managed role definition, name={self.name}')
            logger.debug(f'Data of {self.name} role definition: {defaults}')
            logger.debug(f'Permissions of {self.name} role definition: {permissions}')
        return rd, created

    def allowed_permissions(self, model: Optional[Type[Model]]) -> set[str]:
        from ansible_base.rbac.validators import combine_values, permissions_allowed_for_role

        return combine_values(permissions_allowed_for_role(model))


class ManagedAdminBase(ManagedRoleConstructor):
    description = gettext_noop("Has all permissions to a single {model_name_verbose}")

    def get_permissions(self, apps) -> set[str]:
        """All permissions possible for the associated model"""
        return self.allowed_permissions(self.get_model(apps))


class ManagedActionBase(ManagedRoleConstructor):
    description = gettext_noop("Can take specified action for a single {model_name_verbose}")
    action = None

    def get_permissions(self, apps) -> set[str]:
        """Gives permission for one special action and includes view permission as well"""
        model_name = self.get_model(apps)._meta.model_name
        return {f'view_{model_name}', self.action}


class ManagedReadOnlyBase(ManagedRoleConstructor):
    """Given a certain type this managed role includes all possible view permissions for that type

    The type is defined in the subclass, so this is an abstract class
    """

    description = gettext_noop("Has all viewing related permissions that can be delegated via {model_name_verbose}")

    def get_permissions(self, apps) -> set[str]:
        return {codename for codename in self.allowed_permissions(self.get_model(apps)) if codename.startswith('view')}


class OrganizationMixin:
    model_name = settings.ANSIBLE_BASE_ORGANIZATION_MODEL


class TeamMixin:
    model_name = settings.ANSIBLE_BASE_TEAM_MODEL


# Start concrete shared role definitions


class SystemAuditor(ManagedReadOnlyBase):
    name = gettext_noop("System Auditor")
    description = gettext_noop("Has view permissions to all objects")
    model_name = None


class OrganizationAdmin(OrganizationMixin, ManagedAdminBase):
    name = gettext_noop("Organization Admin")
    description = gettext_noop("Has all permissions to a single organization and all objects inside of it")


class OrganizationMember(OrganizationMixin, ManagedActionBase):
    name = gettext_noop("Organization Member")
    description = gettext_noop("Has member permission to a single organization")
    action = 'member_organization'


class TeamAdmin(TeamMixin, ManagedAdminBase):
    name = gettext_noop("Team Admin")
    description = gettext_noop("Can manage a single team and inherits all role assignments to the team")


class TeamMember(TeamMixin, ManagedActionBase):
    name = gettext_noop("Team Member")
    description = gettext_noop("Inherits all role assignments to a single team")
    action = 'member_team'


# Setup for registry, ultimately exists inside of permission_registry


managed_role_templates = {
    'sys_auditor': SystemAuditor,
    'org_admin': OrganizationAdmin,
    'org_member': OrganizationMember,
    'team_admin': TeamAdmin,
    'team_member': TeamMember,
    # These are not fully functional on their own, but can be easily subclassed
    'admin_base': ManagedAdminBase,
    'action_base': ManagedActionBase,
}


def get_managed_role_constructors(apps, setting_value: dict[str, dict]) -> dict[str, ManagedRoleConstructor]:
    """Constructs managed role definition (instructions for creating a managed role definition)

    from the entries in setting_value, expected to be from settings.ANSIBLE_BASE_MANAGED_ROLE_REGISTRY"""
    ret = {}
    for shortname, role_data in setting_value.items():
        lookup_shortname = role_data.get('shortname', shortname)
        cls = managed_role_templates[lookup_shortname]
        overrides = role_data.copy()
        overrides['template_shortname'] = lookup_shortname
        overrides['shortname'] = shortname
        ret[shortname] = cls(overrides=overrides)
    return ret



from .common import CommonModel, ImmutableCommonModel, NamedCommonModel, UniqueNamedCommonModel
from .immutable import ImmutableModel
from .organization import AbstractOrganization
from .team import AbstractTeam
from .user import AbstractDABUser

__all__ = (
    'AbstractOrganization',
    'AbstractTeam',
    'AbstractDABUser',
    'CommonModel',
    'ImmutableModel',
    'ImmutableCommonModel',
    'NamedCommonModel',
    'UniqueNamedCommonModel',
)



from datetime import datetime, timedelta
from typing import TypedDict

import jwt
from django.conf import settings

from ansible_base.resource_registry.models import service_id


class ResourceServerConfig(TypedDict):
    URL: str
    SECRET_KEY: str
    VALIDATE_HTTPS: bool
    JWT_ALGORITHM: str


def get_resource_server_config() -> ResourceServerConfig:
    defaults = {"JWT_ALGORITHM": "HS256", "VALIDATE_HTTPS": True}
    defaults.update(settings.RESOURCE_SERVER)
    return defaults


def get_service_token(user_id=None, expiration=60, **kwargs):
    config = get_resource_server_config()
    payload = {
        "iss": str(service_id()),
        **kwargs,
    }

    if user_id is not None:
        payload["sub"] = user_id

    if expiration is not None:
        payload["exp"] = datetime.now() + timedelta(seconds=expiration)

    return jwt.encode(payload, config["SECRET_KEY"], config["JWT_ALGORITHM"])



import logging
from contextlib import contextmanager
from typing import Optional, Union
from uuid import UUID

from django.db.models import Model, Q
from django.db.models.signals import m2m_changed, post_delete, post_init, post_save, pre_delete, pre_save
from django.dispatch import Signal

from ansible_base.lib.utils.db import migrations_are_complete
from ansible_base.rbac.caching import compute_object_role_permissions, compute_team_member_roles
from ansible_base.rbac.models import ObjectRole, RoleDefinition, RoleEvaluation, get_evaluation_model
from ansible_base.rbac.permission_registry import permission_registry
from ansible_base.rbac.validators import validate_team_assignment_enabled

logger = logging.getLogger('ansible_base.rbac.triggers')


"""
As the caching module will fill in cached data,
this module shall manage the calling of the caching methods.
Sounds simple, but is actually more complicated that the caching logic itself.
"""


dab_post_migrate = Signal()


def team_ancestor_roles(team):
    """
    Return a queryset of all roles that directly or indirectly grant any form of permission to a team.
    This is generally used when invalidating a team membership for one reason or another.
    This assumes that teams and all team parent models have integer primary keys.
    """
    permission_kwargs = dict(codename=permission_registry.team_permission, object_id=team.id, content_type_id=permission_registry.team_ct_id)
    return set(ObjectRole.objects.filter(permission_partials__in=RoleEvaluation.objects.filter(**permission_kwargs)))


def needed_updates_on_assignment(role_definition, actor, object_role, created=False, giving=True):
    """
    If a user or a team is granted a role or has a role revoked,
    then this returns instructions for what needs to be updated
    returns tuple
        (bool: should update team owners, set: object roles to update)
    """
    # we maintain a list of object roles that we need to update evaluations for
    to_update = set()
    if created:
        to_update.add(object_role)

    has_team_perm = role_definition.permissions.filter(codename=permission_registry.team_permission).exists()

    if actor._meta.model_name == permission_registry.team_model._meta.model_name:
        has_org_member = role_definition.permissions.filter(codename='member_organization').exists()

        # Raise exception if settings prohibits this assignment
        validate_team_assignment_enabled(object_role.content_type, has_team_perm=has_team_perm, has_org_member=has_org_member)

    # If permissions for team are changed. That tends to affect a lot.
    changes_team_owners = False
    if actor._meta.model_name != 'user':
        to_update.update(team_ancestor_roles(actor))
        if not giving:
            # this will delete some permission assignments that will be removed from this relationship
            to_update.update(object_role.descendent_roles())
        changes_team_owners = True

    deleted = False
    if (not giving) and (not (object_role.users.exists() or object_role.teams.exists())):
        # time to delete the object role because it is unused
        if object_role in to_update:
            to_update.remove(object_role)
        deleted = True

    # giving or revoking team permissions may not change the parentage
    # but this will still change what downstream roles grant what permissions
    if (has_team_perm and created) or (giving and changes_team_owners):
        to_update.update(object_role.descendent_roles())

    # actions which can change the team parentage structure
    recompute_teams = bool(has_team_perm and (created or deleted or changes_team_owners))

    return (recompute_teams, to_update)


def update_after_assignment(update_teams, to_update):
    "Call this with the output of needed_updates_on_assignment"
    if update_teams:
        compute_team_member_roles()

    compute_object_role_permissions(object_roles=to_update)


def permissions_changed(instance, action, model, pk_set, reverse, **kwargs):
    if action.startswith('pre_'):
        return
    to_recompute = set(ObjectRole.objects.filter(role_definition=instance).prefetch_related('teams__member_roles'))
    if not to_recompute:
        return
    if reverse:
        raise RuntimeError('Removal of permssions through reverse relationship not supported')

    if action in ('post_add', 'post_remove'):
        if permission_registry.permission_qs.filter(codename=permission_registry.team_permission, pk__in=pk_set).exists():
            for object_role in to_recompute.copy():
                to_recompute.update(object_role.descendent_roles())
            compute_team_member_roles()
        # All team member roles that give this permission through this role need to be updated
        for role in to_recompute.copy():
            for team in role.teams.all():
                for team_role in team.member_roles.all():
                    to_recompute.add(team_role)
    elif action == 'post_clear':
        # unfortunately this does not give us a list of permissions to work with
        # this is slow, not ideal, but will at least be correct
        compute_team_member_roles()
        to_recompute = None  # all
    compute_object_role_permissions(object_roles=to_recompute)


m2m_changed.connect(permissions_changed, sender=RoleDefinition.permissions.through)


def rbac_post_init_set_original_parent(sender, instance, **kwargs):
    """
    connect to post_init signal
    Used to set the original, or
    pre-save parent id (usually organization), so we can later determine if
    the organization field has changed.
    """
    parent_field_name = permission_registry.get_parent_fd_name(instance)
    if parent_field_name is None:
        return
    parent_id_name = f'{parent_field_name}_id'
    if parent_id_name not in instance.__dict__:
        return  # we do not want to conflit with .only usage
    instance.__rbac_original_parent_id = getattr(instance, parent_id_name)


def get_parent_ids(instance) -> list[tuple[Model, Union[int, UUID]]]:
    parent_field_name = permission_registry.get_parent_fd_name(instance)
    if not parent_field_name:
        return []
    parent_cls = permission_registry.get_parent_model(instance)

    if permission_registry.get_parent_fd_name(parent_cls):
        # has another level of model
        parent_obj = getattr(instance, parent_field_name)
        if parent_obj:
            parent_ct = permission_registry.content_type_model.objects.get_for_model(parent_cls)
            return [(parent_ct, parent_obj.pk)] + get_parent_ids(parent_obj)
    else:
        parent_id = getattr(instance, f'{parent_field_name}_id')
        if parent_id:
            parent_ct = permission_registry.content_type_model.objects.get_for_model(parent_cls)
            return [(parent_ct, parent_id)]
    return []


def post_save_update_obj_permissions(instance):
    "Utility method shared by multiple signals"
    # Account for organization roles (and other parent objects), new and old
    parent_gfks = get_parent_ids(instance)

    if hasattr(instance, '__rbac_original_parent_id'):
        parent_cls = permission_registry.get_parent_model(instance)
        parent_ct = permission_registry.content_type_model.objects.get_for_model(parent_cls)
        parent_obj = parent_cls(pk=instance.__rbac_original_parent_id)
        parent_gfks += get_parent_ids(parent_obj)
        parent_gfks.append((parent_ct, instance.__rbac_original_parent_id))
        delattr(instance, '__rbac_original_parent_id')

    if parent_gfks:
        q_exprs = [Q(content_type=parent_ct, object_id=parent_id) for parent_ct, parent_id in parent_gfks]
        q_filter = q_exprs[0]
        for next_q in q_exprs[1:]:
            q_filter |= next_q
        to_update = set(ObjectRole.objects.filter(q_filter))
    else:
        to_update = set()

    # Account for parent team roles of those organization roles
    ancestors = set(ObjectRole.objects.filter(provides_teams__has_roles__in=to_update))
    to_update.update(ancestors)

    # If the actual object changed (created or modified) was a team, any org role
    # that has member_team needs to be updated, and any parent teams that have that role
    if instance._meta.model_name == permission_registry.team_model._meta.model_name:
        compute_team_member_roles()

    if to_update:
        compute_object_role_permissions(object_roles=to_update)


def rbac_pre_save_identify_changes(instance, *args, **kwargs):
    # Exit right away if object does not have any parent objects
    parent_field_name = permission_registry.get_parent_fd_name(instance)
    if parent_field_name is None:
        return

    # The parent object can not have changed if update_fields was given and did not list that field
    update_fields = kwargs.get('update_fields', None)
    if update_fields and not (parent_field_name in update_fields or f'{parent_field_name}_id' in update_fields):
        return

    # If we HAVE to do a query to find out if the parent field has changed then we will here
    if not hasattr(instance, '__rbac_original_parent_id') and instance.pk:
        instance.__rbac_original_parent_id = getattr(type(instance).objects.only('pk').get(pk=instance.pk), f'{parent_field_name}_id')


def rbac_post_save_update_evaluations(instance, created, *args, **kwargs):
    """
    Connect to post_save signal for objects in the permission registry
    If the parent object changes, this rebuilds the cache
    """
    # Exit right away if object does not have any parent objects
    parent_field_name = permission_registry.get_parent_fd_name(instance)
    if parent_field_name is None:
        return

    # If child object is created and parent object has existing ObjectRoles
    # evaluations for the parent object roles need to be added
    if created:
        post_save_update_obj_permissions(instance)
        return

    # The parent object can not have changed if update_fields was given and did not list that field
    update_fields = kwargs.get('update_fields', None)
    if update_fields and not (parent_field_name in update_fields or f'{parent_field_name}_id' in update_fields):
        return

    # Handle the unusual situation where the parent object changes
    current_parent_id = getattr(instance, f'{parent_field_name}_id')
    if hasattr(instance, '__rbac_original_parent_id') and instance.__rbac_original_parent_id != current_parent_id:
        logger.info(f'Object {instance} changed RBAC parent {instance.__rbac_original_parent_id}-->{current_parent_id}')
        post_save_update_obj_permissions(instance)


def team_pre_delete(instance, *args, **kwargs):
    instance.__rbac_stashed_member_roles = list(instance.member_roles.all())


def rbac_post_delete_remove_object_roles(instance, *args, **kwargs):
    """
    Call this when deleting an object to cascade delete its object roles
    Deleting a team can have consequences for the rest of the graph
    """
    if instance._meta.model_name == permission_registry.team_model._meta.model_name:
        indirectly_affected_roles = set()
        indirectly_affected_roles.update(team_ancestor_roles(instance))
        for team_role in instance.__rbac_stashed_member_roles:
            indirectly_affected_roles.update(team_role.descendent_roles())
        compute_team_member_roles()
        compute_object_role_permissions(object_roles=indirectly_affected_roles)

        # Similar to user deletion, clean up any orphaned object roles
        ObjectRole.objects.filter(users__isnull=True, teams__isnull=True).delete()

    ct = permission_registry.content_type_model.objects.get_for_model(instance)
    ObjectRole.objects.filter(content_type=ct, object_id=instance.pk).delete()

    parent_field_name = permission_registry.get_parent_fd_name(instance)
    if parent_field_name:
        # Delete all evaluations from inherited permissions
        get_evaluation_model(instance).objects.filter(content_type_id=ct.id, object_id=instance.pk).delete()


def rbac_post_user_delete(instance, *args, **kwargs):
    """
    After you delete a user, all their permissions should be removed as well
    """
    # Any RoleUserAssignment entries will already be cascade deleted
    # Just clean up any object roles that may be orphaned by this deletion
    ObjectRole.objects.filter(users__isnull=True, teams__isnull=True).delete()


def post_migration_rbac_setup(sender, *args, **kwargs):
    if not migrations_are_complete():
        logger.info('Not running DAB RBAC post_migrate logic because of suspected reverse migration')
        return

    dab_post_migrate.send(sender=sender)

    compute_team_member_roles()
    compute_object_role_permissions()


class TrackedRelationship:
    def __init__(self, cls, role_name):
        self.cls = cls
        self.role_name = role_name
        self.user_relationship = None
        self.team_relationship = None
        self._active_sync_flag = False

    def initialize(self, relationship):
        manager = getattr(self.cls, relationship)
        related_model_name = manager.field.related_model._meta.model_name
        if related_model_name == permission_registry.team_model._meta.model_name:
            self.team_relationship = relationship
            m2m_changed.connect(self.sync_team_to_role, sender=manager.through)
        elif related_model_name == permission_registry.user_model._meta.model_name:
            self.user_relationship = relationship
            m2m_changed.connect(self.sync_user_to_role, sender=manager.through)
        else:
            raise RuntimeError(f'Can only register user or team relationships, obtained {related_model_name}')

    @contextmanager
    def sync_active(self):
        try:
            self._active_sync_flag = True
            yield
        finally:
            self._active_sync_flag = False

    def sync_relationship(self, actor, content_object, giving=True):
        # Exit if role does not apply for the intended model type, for example
        # if user is given "team-member" role to organization, do not add user to the team members
        if content_object._meta.model_name != self.cls._meta.model_name:
            return

        if actor._meta.model_name == permission_registry.team_model._meta.model_name:
            if self.team_relationship is None:
                return
            manager = getattr(content_object, self.team_relationship)
        elif actor._meta.model_name == permission_registry.user_model._meta.model_name:
            if self.user_relationship is None:
                return
            manager = getattr(content_object, self.user_relationship)

        if giving:
            manager.add(actor)
        else:
            manager.remove(actor)

    def _sync_actor_to_role(self, actor_model: type, instance: Model, action: str, pk_set: Optional[set[int]]):
        if self._active_sync_flag:
            return
        if action.startswith('pre_'):
            return
        rd = RoleDefinition.objects.get(name=self.role_name)

        if action in ('post_add', 'post_remove'):
            actor_set = pk_set
        elif action == 'post_clear':
            ct = permission_registry.content_type_model.objects.get_for_model(instance)
            role = ObjectRole.objects.get(object_id=instance.pk, content_type=ct, role_definition=rd)
            if actor_model._meta.model_name == 'team':
                actor_set = set(role.teams.values_list('id', flat=True))
            else:
                actor_set = set(role.users.values_list('id', flat=True))

        giving = bool(action == 'post_add')
        for actor in actor_model.objects.filter(pk__in=actor_set):
            rd.give_or_remove_permission(actor, instance, giving=giving, sync_action=True)

    def sync_team_to_role(self, instance: Model, action: str, model: type, pk_set: Optional[set[int]], reverse: bool, **kwargs):
        if not reverse:
            self._sync_actor_to_role(permission_registry.team_model, instance, action, pk_set)
        else:
            for pk in pk_set:
                self._sync_actor_to_role(permission_registry.team_model, model(pk=pk), action, {instance.pk})

    def sync_user_to_role(self, instance: Model, action: str, model: type, pk_set: Optional[set[int]], reverse: bool, **kwargs):
        if not reverse:
            self._sync_actor_to_role(permission_registry.user_model, instance, action, pk_set)
        else:
            for pk in pk_set:
                self._sync_actor_to_role(permission_registry.user_model, model(pk=pk), action, {instance.pk})


def connect_rbac_signals(cls):
    if cls._meta.model_name == permission_registry.team_model._meta.model_name:
        pre_delete.connect(team_pre_delete, sender=cls, dispatch_uid='stash-team-roles-before-delete')

    post_init.connect(rbac_post_init_set_original_parent, sender=cls, dispatch_uid='permission-registry-save-prior-parent')
    pre_save.connect(rbac_pre_save_identify_changes, sender=cls, dispatch_uid='permission-registry-pre-save')
    post_save.connect(rbac_post_save_update_evaluations, sender=cls, dispatch_uid='permission-registry-post-save')
    post_delete.connect(rbac_post_delete_remove_object_roles, sender=cls, dispatch_uid='permission-registry-post-delete')



import logging
import os

from django import template
from django.conf import settings
from django.utils.safestring import mark_safe

logger = logging.getLogger("ansible_base.lib.templatetags.util")
register = template.Library()


@register.simple_tag
def inline_file(relative_path, is_safe=False, fatal=False):
    path = os.path.join(settings.BASE_DIR, relative_path)
    try:
        with open(path, 'r') as file:
            contents = file.read()
            if is_safe:
                contents = mark_safe(contents)
            return contents
    except Exception:
        logger.exception(f"Failed to read file {path}")
        if fatal:
            raise



import logging

from django.urls import re_path

from ansible_base.jwt_consumer.apps import JwtConsumerConfig
from ansible_base.jwt_consumer.views import PlatformUIRedirectView

logger = logging.getLogger('ansible_base.jwt_consumer.urls')

# This is a special case because the application has to include this in a very specific location
# in order for the redirect to be picked up.
# Therefore we will not add it to our standard api_urls/api_root_urls/root_url variables.

app_name = JwtConsumerConfig.label
urlpatterns = [
    re_path(r'', PlatformUIRedirectView.as_view()),
]



import logging

from django.core.management import call_command
from django.core.management.base import BaseCommand

from ansible_base.oauth2_provider.models import OAuth2AccessToken, OAuth2RefreshToken


class Command(BaseCommand):
    def init_logging(self):
        log_levels = dict(enumerate([logging.ERROR, logging.INFO, logging.DEBUG, 0]))
        self.logger = logging.getLogger('ansible_base.oauth2_provider.commands.cleanup_tokens')
        self.logger.setLevel(log_levels.get(self.verbosity, 0))
        handler = logging.StreamHandler(stream=self.stream)
        handler.setFormatter(logging.Formatter('%(message)s'))
        self.logger.addHandler(handler)
        self.logger.propagate = False

    def execute(self, *args, **options):
        self.verbosity = int(options.get('verbosity', 1))
        self.stream = options.get('stderr')
        self.init_logging()
        total_accesstokens = OAuth2AccessToken.objects.all().count()
        total_refreshtokens = OAuth2RefreshToken.objects.all().count()
        call_command("cleartokens")
        self.logger.info("Expired OAuth 2 Access Tokens deleted: {}".format(total_accesstokens - OAuth2AccessToken.objects.all().count()))
        self.logger.info("Expired OAuth 2 Refresh Tokens deleted: {}".format(total_refreshtokens - OAuth2RefreshToken.objects.all().count()))



# Generated by Django 4.2.11 on 2024-06-06 22:46

import ansible_base.oauth2_provider.models.access_token
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_oauth2_provider', '0003_remove_oauth2application_logo_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='scope',
            field=models.CharField(default='write', help_text="Allowed scopes, further restricts user's permissions. Must be a simple space-separated string with allowed scopes ['read', 'write'].", max_length=32, validators=[ansible_base.oauth2_provider.models.access_token.validate_scope]),
        ),
    ]



# Generated by Django 4.2.8 on 2024-04-13 21:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dab_activitystream', '0002_alter_entry_created_by'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='entry',
            options={'ordering': ['id'], 'verbose_name_plural': 'Entries'},
        ),
    ]



"""Organization models."""

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from .common import NamedCommonModel


class AbstractTeam(NamedCommonModel):
    """
    An abstract base class for teams.
    A team groups users that work on the same resources, so that permissions can be assigned efficiently.
    """

    class Meta:
        abstract = True
        unique_together = [('organization', 'name')]
        ordering = ('organization__name', 'name')

    description = models.TextField(
        null=False,
        blank=True,
        default="",
        help_text=_("The team description."),
    )

    organization = models.ForeignKey(
        settings.ANSIBLE_BASE_ORGANIZATION_MODEL,
        blank=False,
        null=False,
        on_delete=models.CASCADE,
        related_name="teams",
        help_text=_("The organization of this team."),
    )



from rest_framework.viewsets import ModelViewSet

from ansible_base.authentication.models import AuthenticatorMap
from ansible_base.authentication.serializers import AuthenticatorMapSerializer
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView


class AuthenticatorMapViewSet(AnsibleBaseDjangoAppApiView, ModelViewSet):
    """
    API endpoint that allows authenticator maps to be viewed or edited.
    """

    queryset = AuthenticatorMap.objects.all().order_by("id")
    serializer_class = AuthenticatorMapSerializer



# Generated by Django 4.2.8 on 2024-03-11 18:57

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0003_alter_authenticatormap_authenticator'),
    ]

    operations = [
        migrations.RenameField(
            model_name='authenticator',
            old_name='created_on',
            new_name='created',
        ),
        migrations.RenameField(
            model_name='authenticator',
            old_name='modified_on',
            new_name='modified',
        ),
        migrations.RenameField(
            model_name='authenticatormap',
            old_name='created_on',
            new_name='created',
        ),
        migrations.RenameField(
            model_name='authenticatormap',
            old_name='modified_on',
            new_name='modified',
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, help_text='The date/time this resource was created'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='created',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, help_text='The date/time this resource was created'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created'),
        ),
    ]



import csv
import logging
from dataclasses import dataclass
from typing import Optional, Sequence
from urllib.parse import urlparse

from crum import get_current_request
from django.http import StreamingHttpResponse
from django.urls import reverse as django_reverse
from rest_framework.reverse import reverse as drf_reverse

from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.lib.utils.response')


class CSVBuffer:
    """An object that implements just the write method of the file-like Protocol
    # NOTE: cannot use io.StringIO because its write returns length instead of text
    # we need the text to be immediatelly written to the streaming response.
    """

    def write(self, value):
        return value


@dataclass
class CSVStreamResponse:
    """Streams a CSV from any sequence e.g: queryset, generator"""

    lines: Sequence[Sequence[str]]  # can be a generator that yields tuple[str]
    filename: Optional[str] = None
    content_type: str = "text/event-stream"
    headers: Optional[dict] = None

    def stream(self):
        writer = csv.writer(CSVBuffer())
        headers = {"Cache-Control": "no-cache"} if self.headers is None else self.headers
        if self.filename:  # pragma: no cover
            headers["Content-Disposition"] = f"attachment; filename={self.filename}"

        return StreamingHttpResponse((writer.writerow(line) for line in self.lines), status=200, content_type=self.content_type, headers=headers)


def get_fully_qualified_url(view_name: str, *args, **kwargs) -> str:
    '''
    Returns a fully qualified URL from the setting FRONT_END_URL or a DRF reverse if the setting is undefined
    NOTE: a DRF reverse could be relative or not depending on the request coming in
    '''
    front_end_url = get_setting('FRONT_END_URL', None)
    if front_end_url:
        url = drf_reverse(view_name, *args, **kwargs)
        url_pieces = urlparse(url)
        return f"{front_end_url.rstrip('/')}{url_pieces.path}"

    if not kwargs.get('request'):
        kwargs['request'] = get_current_request()

    return drf_reverse(view_name, *args, **kwargs)


def get_relative_url(view_name: str, *args, **kwargs) -> str:
    '''
    Returns a relative URL to the specified view_name
    '''
    return django_reverse(view_name, *args, **kwargs)



from django.apps import AppConfig

import ansible_base.lib.checks  # noqa: F401 - register checks


class AuthenticationConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.authentication'
    label = 'dab_authentication'
    verbose_name = 'Pluggable Authentication'



from collections import OrderedDict

from django.utils.translation import gettext_lazy as _
from rest_framework import permissions
from rest_framework.response import Response

from ansible_base.lib.utils.response import get_relative_url
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView


class ApiOAuthAuthorizationRootView(AnsibleBaseDjangoAppApiView):
    permission_classes = (permissions.AllowAny,)
    name = _("API OAuth 2 Authorization Root")
    versioning_class = None
    swagger_topic = 'Authentication'

    def get(self, request, format=None):
        data = OrderedDict()
        data['authorize'] = get_relative_url('authorize')
        data['revoke_token'] = get_relative_url('revoke-token')
        data['token'] = get_relative_url('token')
        return Response(data)



import importlib
import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.db.utils import IntegrityError
from django.http import HttpResponseNotFound
from social_core.utils import setting_name
from social_django.models import Association, Code, Nonce, Partial
from social_django.storage import BaseDjangoStorage
from social_django.strategy import DjangoStrategy

from ansible_base.authentication.authenticator_plugins.utils import generate_authenticator_slug, get_authenticator_class, get_authenticator_plugins
from ansible_base.authentication.models import Authenticator, AuthenticatorUser
from ansible_base.lib.utils.response import get_fully_qualified_url

logger = logging.getLogger('ansible_base.authentication.social_auth')


SOCIAL_AUTH_PIPELINE_FAILED_STATUS = "pipeline-failed"


class AuthenticatorStorage(BaseDjangoStorage):
    user = AuthenticatorUser
    nonce = Nonce
    association = Association
    code = Code
    partial = Partial

    @classmethod
    def is_integrity_error(cls, exception):
        return exception.__class__ is IntegrityError


class AuthenticatorStrategy(DjangoStrategy):
    def __init__(self, storage, request=None, tpl=None):
        super().__init__(storage, request, tpl)
        self.settings = {}
        fq_function_name = getattr(settings, 'ANSIBLE_BASE_SOCIAL_AUTH_STRATEGY_SETTINGS_FUNCTION', None)
        if fq_function_name:
            logger.debug(f"Attempting to load social settings from {fq_function_name}")
            try:
                module_name, _, function_name = fq_function_name.rpartition('.')
                the_function = getattr(importlib.import_module(module_name), function_name)
                self.settings = the_function()
            except Exception as e:
                logger.error(f"Failed to run {fq_function_name} to get additional settings: {e}")

    # override setting to pass the backend to get_setting
    def setting(self, name, default=None, backend=None):
        names = [setting_name(name), name]
        if backend:
            names.insert(0, setting_name(backend.name, name))
        for name in names:
            try:
                return self.get_setting(name, backend)
            except (AttributeError, KeyError):
                pass
        return default

    # load the authenticator setting from the database object.
    def get_setting(self, name, backend):
        # try to load the value from the db.
        if backend and hasattr(backend, 'database_instance'):
            value = backend.database_instance.configuration.get(name, None)
            if value is not None:
                return value

            # next check the ADDITIONAL_UNVERIFIED_ARGS
            additional_args = backend.database_instance.configuration.get('ADDITIONAL_UNVERIFIED_ARGS', {})
            value = additional_args.get(name, None)
            if value is not None:
                return value

        # See if we have the setting ourselves
        if self.settings.get(name, None):
            return self.settings.get(name)

        # fall back to settings module
        return super().get_setting(name)

    def get_backends(self):
        """Return configured backends"""
        return get_authenticator_plugins()

    def get_backend_class(self, name):
        """Return a configured backend class"""
        # TODO: This can raise an exception if the back end is fubar
        return get_authenticator_class(name)

    def get_backend(self, slug, redirect_uri=None, *args, **kwargs):
        """Add the database instance arg into the social auth backend."""

        db_instance = Authenticator.objects.get(slug=slug)
        Backend = self.get_backend_class(db_instance.type)

        kwargs["database_instance"] = db_instance
        kwargs["redirect_uri"] = redirect_uri
        args = (self,) + args

        return Backend(
            *args,
            **kwargs,
        )

    def session_set(self, name, value):
        # Social auth tries to set auth_user.provider as a value in the session. Since
        # we've set provider to be a foreign key to the Authenticator model, this can
        # receive an Authenticator object, which isn't json serializeable and causes the
        # session to break.
        if isinstance(value, models.Model):
            value = str(value)
        return super().session_set(name, value)

    def create_user(self, *args, **kwargs):
        # In the social pipeline we still want to call social_core.pipeline.user.create_user
        #     because it will pull in the user fields if set.
        # However, we want to be able to connect to existing Users which social auth does not handle well.
        # This is a short circuit to return an already created user to appease social auth login in our model.
        try:
            # Return the existing user if it already exists.
            return get_user_model().objects.get(username=kwargs.get('username', None))
        except get_user_model().DoesNotExist:
            # Call the parent class if the User has not already been created.
            return self.storage.user.create_user(*args, **kwargs)


class AuthenticatorConfigTestStrategy(AuthenticatorStrategy):
    def __init__(self, storage, request=None, tpl=None, additional_settings={}):
        super().__init__(storage, request, tpl)
        self.settings.update(additional_settings)


class SocialAuthMixin:
    configuration_encrypted_fields = []
    logger = None
    groups_claim = "Group"

    def __init__(self, *args, **kwargs):
        # social auth expects the first arg to be a strategy instance. Since this has
        # to be instantiated outside of social auth, make sure that the strategy arg
        # is present.
        args = self.ensure_strategy_in_args(args)
        self.database_instance = kwargs.pop("database_instance", None)
        super().__init__(*args, **kwargs)
        self.set_logger(self.logger)

    def start(self):
        # This will be run on the /login call and we want to return a 404 if the authenticator is not enabled.
        if not self.database_instance.enabled:
            logger.error(f"Authentication attempted with disabled authenticator {self.database_instance.name}")
            return HttpResponseNotFound()
        return super().start()

    @property
    def name(self):
        return str(self.database_instance.slug)

    def get_user_groups(self, extra_groups=None):
        """
        Receives the user object that .authenticate returns.
        """
        return []

    def ensure_strategy_in_args(self, args):
        if len(args) == 0:
            args = (AuthenticatorStrategy(storage=AuthenticatorStorage()),)

        return args


class SocialAuthValidateCallbackMixin:
    def validate(self, serializer, data):
        # if we have an instance already and we didn't get a configuration parameter we are just updating other fields and can return
        if serializer.instance and 'configuration' not in data:
            return data

        configuration = data['configuration']
        if not configuration.get('CALLBACK_URL', None):
            if not serializer.instance:
                data["slug"] = generate_authenticator_slug(data.get("slug"))
                slug = data["slug"]
            else:
                slug = serializer.instance.slug

            configuration['CALLBACK_URL'] = get_fully_qualified_url('social:complete', kwargs={'backend': slug})

        return data


def create_user_claims_pipeline(*args, backend, response, **kwargs):
    from ansible_base.authentication.utils.claims import update_user_claims

    groups_claim = backend.groups_claim if backend.groups_claim is not None else "Group"

    extra_groups = response[groups_claim] if groups_claim in response else []
    user = update_user_claims(kwargs["user"], backend.database_instance, backend.get_user_groups(extra_groups))
    if user is None:
        return SOCIAL_AUTH_PIPELINE_FAILED_STATUS



import logging
from typing import Optional

from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from ansible_base.lib.utils.response import get_relative_url
from ansible_base.resource_registry.models import Resource, ResourceType

logger = logging.getLogger('ansible_base.serializers')


class ResourceDataField(serializers.JSONField):
    """
    Inspects the content object. If it has a managed serializer,
    serialize the data using it.
    """

    def to_representation(self, resource):
        if serializer := resource.content_type.resource_type.serializer_class:
            return serializer(resource.content_object).data
        return {}

    def to_internal_value(self, data):
        data = super().to_internal_value(data)
        return {self.field_name: data}


class ResourceListSerializer(serializers.ModelSerializer):
    has_serializer = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()
    resource_type = serializers.CharField(required=False)
    resource_data = ResourceDataField(source="*", write_only=True)

    class Meta:
        model = Resource
        read_only_fields = [
            "object_id",
            "name",
        ]
        fields = [
            "object_id",
            "name",
            "ansible_id",
            "service_id",
            "is_partially_migrated",
            "resource_type",
            "has_serializer",
            "resource_data",
            "url",
        ]

    def get_url(self, obj) -> str:
        # conversion to string is done to satisfy type checking and OpenAPI schema generator
        return get_relative_url('resource-detail', kwargs={"ansible_id": obj.ansible_id})

    def get_has_serializer(self, obj) -> bool:
        return bool(obj.content_type.resource_type.get_resource_config().managed_serializer)

    # update ansible ID
    def update(self, instance, validated_data):
        resource_type = instance.content_type.resource_type
        if not resource_type.can_be_managed:
            raise serializers.ValidationError({"resource_type": _(f"Resource type: {resource_type.name} cannot be managed by Resources.")})

        instance.update_resource(
            validated_data.get("resource_data", {}),
            ansible_id=validated_data.get("ansible_id"),
            is_partially_migrated=validated_data.get("is_partially_migrated"),
            service_id=validated_data.get("service_id"),
            partial=self.partial,
        )
        instance.refresh_from_db()
        return instance

    # allow setting ansible ID at create time
    def create(self, validated_data):
        try:
            if not validated_data["resource_type"]:
                raise serializers.ValidationError({"resource_type": _("This field is required for resource creation.")})

            resource_type = ResourceType.objects.get(name=validated_data["resource_type"])
            if not resource_type.can_be_managed:
                raise serializers.ValidationError({"resource_type": _(f"Resource type: {resource_type.name} cannot be managed by Resources.")})

            return Resource.create_resource(
                resource_type,
                validated_data["resource_data"],
                ansible_id=validated_data.get("ansible_id"),
                service_id=validated_data.get("service_id"),
                is_partially_migrated=validated_data.get("is_partially_migrated", False),
            )

        except ResourceType.DoesNotExist:
            raise serializers.ValidationError({"resource_type": _(f"Resource type: {validated_data['resource_type']} does not exist.")})


class ResourceSerializer(ResourceListSerializer):
    additional_data = serializers.SerializerMethodField()
    resource_data = ResourceDataField(source="*")

    class Meta:
        model = ResourceListSerializer.Meta.model
        read_only_fields = ResourceListSerializer.Meta.read_only_fields
        fields = ResourceListSerializer.Meta.fields + [
            "additional_data",
        ]

    def get_additional_data(self, obj):
        if serializer := obj.content_type.resource_type.serializer_class:
            if serializer.ADDITIONAL_DATA_SERIALIZER is not None:
                return serializer.ADDITIONAL_DATA_SERIALIZER(obj.content_object).data

        return None


class ResourceTypeSerializer(serializers.ModelSerializer):
    shared_resource_type = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = ResourceType
        fields = ["id", "name", "externally_managed", "shared_resource_type", "url"]

    def get_shared_resource_type(self, obj) -> Optional[str]:
        if serializer := obj.get_resource_config().managed_serializer:
            return serializer.RESOURCE_TYPE
        else:
            return None

    def get_url(self, obj) -> str:
        return get_relative_url('resourcetype-detail', kwargs={"name": obj.name})


class UserAuthenticationSerializer(serializers.Serializer):
    username = serializers.CharField()
    password = serializers.CharField()



from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import TextField
from django.db.models.fields.related_descriptors import ForwardOneToOneDescriptor, ReverseOneToOneDescriptor
from django.db.models.functions import Cast
from django.db.models.sql import AND
from django.db.models.sql.where import WhereNode

from ansible_base.resource_registry.models import Resource


class CustomForwardOneToOneDescriptor(ForwardOneToOneDescriptor):
    def get_queryset(self, **hints):
        return self.field.remote_field.model._base_manager.db_manager(hints=hints).filter(content_type=ContentType.objects.get_for_model(self.field.model))

    def get_prefetch_queryset(self, instances, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        queryset._add_hints(instance=instances[0])

        query = models.Q.create(
            [
                ("object_id__in", {str(obj.pk) for obj in instances}),
            ]
        )

        queryset = queryset.filter(query)

        # I can't find documentation for this anywhere, but this is the best I can come
        # up with for what this tuple is returning:
        # [0]: queryset for the related table
        # [1]: function that returns the value for the related table that needs to match the
        #      current table
        # [2]: function that returns the value for the current table that needs to match the
        #      related table
        # [3]: unknown... but this breaks when set to False.
        # [4]: name of the field that we're joining on the current table
        # [5]: unknown...
        # I think that the queryset is merged on the prefetched queryset based on whether the
        # results of the functions in [1] and [2] match.
        return (
            queryset,
            lambda relobj: relobj.object_id,
            lambda obj: str(obj.pk),
            True,
            self.field.attname,
            False,
        )


class AnsibleResourceField(models.ForeignObject):
    """
    This field creates a reverse one to one relationship to the Resource model
    without adding any new columns to the model. It supports `.select_related`
    and `.<resource_field_name>` lookups the same way that a OneToOneField
    does.
    """

    related_accessor_class = ReverseOneToOneDescriptor
    forward_related_accessor_class = CustomForwardOneToOneDescriptor

    def __init__(self, primary_key_field, **kwargs):
        model = Resource
        from_fields = [primary_key_field]
        to_fields = ["object_id"]

        super().__init__(
            model,
            on_delete=models.DO_NOTHING,
            from_fields=from_fields,
            to_fields=to_fields,
            null=True,
            blank=True,
            editable=False,
            # This informs DRF to exclude the field from serialization by default.
            # It can still be serialized, but the user will have to explicitly
            # declare a serializer to use.
            serialize=False,
            **kwargs,
        )

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        # override the default to always make it private
        # this ensures that no additional columns are created
        super().contribute_to_class(cls, name, private_only=True, **kwargs)

    def get_joining_columns(self, reverse_join=False):
        """
        This method normally provides a tuple of tuples, where each pair contains
        the columns to join the two related models on. This returns something like
        "dab_resource_registry_resource"."content_type_id" and
        "aap_gateway_api_organization"."id". The problem here is that the resource
        registry holds a generic foreign key, which is a different column type than
        the local model and Postgres doesn't permit joins across columns of different
        data types. To solve this, we're returning an empty tuple here to disable
        the default join, and instead handling the join in get_extra_restriction().
        """
        return tuple()

    def get_extra_restriction(self, alias, remote_alias):
        """
        Return a pair condition used for joining and subquery pushdown. The
        condition is something that responds to as_sql(compiler, connection)
        method.

        This function casts the local pk field to a text field so that it can
        be joined with Resource.object_id and adds an additional join condition
        for content_type.
        """

        content_type_field = self.remote_field.model._meta.get_field("content_type")
        object_id_field = self.remote_field.model._meta.get_field("object_id")
        pk_field = self.model._meta.get_field("id")
        contenttype_pk = self.get_content_type().pk

        # cast the local pk to a text field
        # resolves to sql: "resource_model"."pk" :: text = "dab_resource_registry_resource"."object_id"
        object_id_lookup = self.get_lookup("exact")(
            Cast(pk_field.get_col(remote_alias), output_field=TextField()),
            object_id_field.get_col(alias),
        )

        # resolves to sql: "dab_resource_registry_resource"."content_type_id" = resource model content type
        content_type_id_lookup = self.get_lookup("exact")(content_type_field.get_col(alias), contenttype_pk)

        return WhereNode([object_id_lookup, content_type_id_lookup], connector=AND)

    def get_extra_descriptor_filter(self, instance):
        """
        Return an extra filter condition for related object fetching when
        user does 'instance.fieldname', that is the extra filter is used in
        the descriptor of the field.

        The filter should be either a dict usable in .filter(**kwargs) call or
        a Q-object. The condition will be ANDed together with the relation's
        joining columns.

        A parallel method is get_extra_restriction() which is used in
        JOIN and subquery conditions.
        """
        return {"content_type": self.get_content_type()}

    def get_content_type(self):
        """
        Return the content type associated with this field's model.
        """
        return ContentType.objects.get_for_model(self.model)



import base64
import binascii
import re
import secrets
from urllib.parse import urlparse, urlunsplit

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.x509 import load_pem_x509_certificate
from django.core.exceptions import ValidationError as LowLevelValidationError
from django.core.validators import URLValidator
from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import ValidationError

VALID_STRING = _('Must be a valid string')


def validate_url_list(urls: list, schemes: list = ['https'], allow_plain_hostname: bool = False) -> None:
    if type(urls) is not list:
        raise ValidationError("Must be a list of urls")
    errors = []
    for a_url in urls:
        if type(a_url) is not str:
            errors.append(f"{a_url} must be a valid url")
            continue
        try:
            validate_url(a_url, schemes=schemes, allow_plain_hostname=allow_plain_hostname)
        except ValidationError:
            errors.append(f"{a_url} is invalid")
    if errors:
        raise ValidationError(', '.join(errors))


def validate_url(url: str, schemes: list = ['https'], allow_plain_hostname: bool = False) -> None:
    if type(url) is not str:
        raise ValidationError(VALID_STRING)
    if allow_plain_hostname:
        # The default validator will not allow names like https://junk so, if we are ok with simple hostnames we are going to munge up the URL for the validator
        try:
            url_parts = urlparse(url)
        except ValueError as e:
            raise ValidationError(str(e)) from e

        # Determine the user_info part of the URL
        user_info = ''
        if url_parts.username:
            user_info = url_parts.username
        if url_parts.password:
            user_info = f'{user_info}:{url_parts.password}'
        if user_info:
            user_info = f"{user_info}@"

        # Check for a valid port number
        try:
            url_parts.port
        except ValueError as e:
            raise ValidationError(str(e)) from e

        if url_parts.hostname and '.' not in url_parts.hostname:
            if '[' in url:
                # https://www.rfc-editor.org/rfc/rfc2732
                hostname = f'[{url_parts.hostname}]'
            else:
                hostname = f'{url_parts.hostname}.localhost'
            port = f':{url_parts.port}' if url_parts.port else ''
            netloc = f"{user_info}{hostname}{port}"
            # Reconstruct and override the URL with a valid hostname
            url = urlunsplit([url_parts.scheme, netloc, url_parts.path, url_parts.query, url_parts.fragment])

    validator = URLValidator(schemes=schemes)
    try:
        validator(url)
    except LowLevelValidationError as e:
        raise ValidationError(e.message)


def validate_cert_with_key(public_cert_string, private_key_string):
    # Returns:
    # None if one of the parameters wasn't set
    # False if we failed to load an item (should be pre-tried by your serializer)
    # A ValidationError exception if the key/value don't match
    # True if everything checks out, meaning that the certificate and private key form a valid keypair

    if not private_key_string or not public_cert_string:
        return None

    private_key = None
    public_cert = None
    try:
        private_key = serialization.load_pem_private_key(bytes(private_key_string, "UTF-8"), password=None)
        public_cert = load_pem_x509_certificate(bytes(public_cert_string, "UTF-8"))
    except Exception:
        return False

    # Generate nonce for keypair verification
    nonce = secrets.token_bytes(64)
    signature = private_key.sign(nonce, padding.PKCS1v15(), public_cert.signature_hash_algorithm)

    try:
        # We have both pieces of the puzzle, lets make sure they interlock;
        #   do so by verifying the nonce we just signed can be verified by the provided certificate
        public_cert.public_key().verify(
            signature,
            nonce,
            # Depends on the algorithm used to create the certificate
            padding.PKCS1v15(),
            public_cert.signature_hash_algorithm,
        )
    except InvalidSignature:
        raise ValidationError(_("The certificate and private key do not match"))
    except Exception as e:
        error = _("Unable to validate SP cert and key")
        if hasattr(e, 'message'):
            error = f"{error}: {e.message}"
        else:
            error = f"{error}: {e.__class__.__name__}"
        raise ValidationError(error)

    return True


def validate_image_data(data: str) -> None:
    # in case we are passed an empty string, we can skip validation
    if not data:
        return None

    CUSTOM_LOGO_RE = re.compile(r'^data:image/(?:png|jpeg|gif);base64,([A-Za-z0-9+/=]+?)$')

    match = CUSTOM_LOGO_RE.match(data)
    if not match:
        raise ValidationError(_("Invalid format for custom logo. Must be a data URL with a base64-encoded GIF, PNG or JPEG image."))
    b64data = match.group(1)
    try:
        base64.b64decode(b64data)
    except (TypeError, binascii.Error):
        raise ValidationError(_("Invalid base64-encoded data in data URL."))


def to_python_boolean(value, allow_none=False):
    value = str(value)
    if value.lower() in ('true', '1', 't'):
        return True
    elif value.lower() in ('false', '0', 'f'):
        return False
    elif allow_none and (value is None or value.lower() in ('none', 'null')):
        return None
    else:
        raise ValueError(_(u'Unable to convert "%s" to boolean') % value)



from typing import Optional

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.db.models.functions import Cast
from django.db.models.query import QuerySet
from rest_framework.serializers import ValidationError

from ansible_base.rbac import permission_registry
from ansible_base.rbac.models import DABPermission, RoleDefinition, get_evaluation_model
from ansible_base.rbac.validators import validate_codename_for_model

"""
RoleEvaluation or RoleEvaluationUUID models are the authority for permission evaluations,
meaning, determining whether a user has a permission to an object.

Methods needed for producing querysets (of objects a user has a permission to
or users that have a permission to an object) or making single evaluations
are defined on the RoleEvaluation model.

This module has logic to attach those evaluation methods to the external
models in an app using these RBAC internals.
"""


def has_super_permission(user, full_codename=None) -> bool:
    "Analog to has_obj_perm but only evaluates to True if user has this permission system-wide"
    if isinstance(user, AnonymousUser):
        return False

    if user._meta.model_name == permission_registry.user_model._meta.model_name:
        # Super permission flags only exist for users, teams can use global roles
        for super_flag in settings.ANSIBLE_BASE_BYPASS_SUPERUSER_FLAGS:
            if getattr(user, super_flag):
                return True  # User has admin flag like is_superuser
        if full_codename:
            for action, super_flag in settings.ANSIBLE_BASE_BYPASS_ACTION_FLAGS.items():
                if full_codename == action and getattr(user, super_flag):
                    return True  # User has action-specific flag like is_platform_auditor
    elif user._meta.model_name != permission_registry.team_model._meta.model_name:
        raise RuntimeError(f'Evaluation methods are for users or teams, got {user._meta.model_name}: {user}')

    if full_codename:
        if full_codename in user.singleton_permissions():
            return True  # User has system role for this action
    return False


def bound_singleton_permissions(self):
    "Method attached to User model as singleton_permissions"
    if not hasattr(self, '_singleton_permissions') or bound_singleton_permissions._team_clear_signal:
        # values_list will make the return type set[str]
        permission_qs = DABPermission.objects.values_list('codename', flat=True)
        self._singleton_permissions = RoleDefinition.user_global_permissions(self, permission_qs=permission_qs)
        bound_singleton_permissions._team_clear_signal = False
    return self._singleton_permissions


bound_singleton_permissions._team_clear_signal = False


class BaseEvaluationDescriptor:
    """
    Descriptors have to be used to attach what are effectively a @classmethod
    to an external model, like MyModel.accessible_objects(u, 'view_mymodel')
    because this how we obtain a reference to MyModel
    """

    def __init__(self, cls):
        self.cls = cls


class AccessibleObjectsDescriptor(BaseEvaluationDescriptor):
    def __call__(self, actor, codename: str = 'view', queryset: Optional[QuerySet] = None) -> QuerySet:
        if queryset is None:
            queryset = self.cls.objects.all()
        if isinstance(actor, AnonymousUser):
            return queryset.model.objects.none()
        if codename == 'view' and ('view' not in self.cls._meta.default_permissions):
            # Model does not track view permissions
            return queryset
        full_codename = validate_codename_for_model(codename, self.cls)
        if actor._meta.model_name == 'user' and has_super_permission(actor, full_codename):
            return queryset
        return get_evaluation_model(self.cls).accessible_objects(self.cls, actor, full_codename, queryset=queryset)


class AccessibleIdsDescriptor(BaseEvaluationDescriptor):
    def __call__(self, actor, codename: str = 'view', content_types=None, cast_field=None) -> QuerySet:
        full_codename = validate_codename_for_model(codename, self.cls)
        if isinstance(actor, AnonymousUser):
            return self.cls.objects.none().values_list()
        if actor._meta.model_name == 'user' and has_super_permission(actor, full_codename):
            if cast_field is None:
                return self.cls.objects.values_list('id', flat=True)
            else:
                return self.cls.objects.values_list(Cast('id', output_field=cast_field), flat=True)
        return get_evaluation_model(self.cls).accessible_ids(self.cls, actor, full_codename, content_types=content_types, cast_field=cast_field)


def bound_has_obj_perm(self, obj, codename) -> bool:
    if not permission_registry.is_registered(obj):
        raise ValidationError(f'Object of {obj._meta.model_name} type is not registered with DAB RBAC')
    full_codename = validate_codename_for_model(codename, obj)
    if has_super_permission(self, full_codename):
        return True
    return get_evaluation_model(obj).has_obj_perm(self, obj, full_codename)


def connect_rbac_methods(cls):
    cls.add_to_class('access_qs', AccessibleObjectsDescriptor(cls))
    cls.add_to_class('access_ids_qs', AccessibleIdsDescriptor(cls))



# Copyright 2024 Red Hat, Inc.
# Copyright (c) 2015, Rob Golding. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Rob Golding, nor the names of its contributors may
#       be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# NOTE(cutwater): Disabling linters to keep diff as little as possible.
# fmt: off
# isort: skip_file
# flake8: noqa
import logging
from io import StringIO

from pyrad.packet import AccessRequest, AccessAccept, AccessReject
from pyrad.client import Client, Timeout
from pyrad.dictionary import Dictionary

from django.conf import settings as global_settings
#Handle custom user models
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins._radiusauth')

User = get_user_model()

DICTIONARY = u"""
ATTRIBUTE   User-Name       1   string
ATTRIBUTE   User-Password       2   string
ATTRIBUTE   CHAP-Password       3   octets
ATTRIBUTE   NAS-IP-Address      4   ipaddr
ATTRIBUTE   NAS-Port        5   integer
ATTRIBUTE   Service-Type        6   integer
ATTRIBUTE   Framed-Protocol     7   integer
ATTRIBUTE   Framed-IP-Address   8   ipaddr
ATTRIBUTE   Framed-IP-Netmask   9   ipaddr
ATTRIBUTE   Framed-Routing      10  integer
ATTRIBUTE   Filter-Id       11  string
ATTRIBUTE   Framed-MTU      12  integer
ATTRIBUTE   Framed-Compression  13  integer
ATTRIBUTE   Login-IP-Host       14  ipaddr
ATTRIBUTE   Login-Service       15  integer
ATTRIBUTE   Login-TCP-Port      16  integer
ATTRIBUTE   Reply-Message       18  string
ATTRIBUTE   Callback-Number     19  string
ATTRIBUTE   Callback-Id     20  string
ATTRIBUTE   Framed-Route        22  string
ATTRIBUTE   Framed-IPX-Network  23  ipaddr
ATTRIBUTE   State           24  octets
ATTRIBUTE   Class           25  octets
ATTRIBUTE   Vendor-Specific     26  octets
ATTRIBUTE   Session-Timeout     27  integer
ATTRIBUTE   Idle-Timeout        28  integer
ATTRIBUTE   Termination-Action  29  integer
ATTRIBUTE   Called-Station-Id   30  string
ATTRIBUTE   Calling-Station-Id  31  string
ATTRIBUTE   NAS-Identifier      32  string
ATTRIBUTE   Proxy-State     33  octets
ATTRIBUTE   Login-LAT-Service   34  string
ATTRIBUTE   Login-LAT-Node      35  string
ATTRIBUTE   Login-LAT-Group     36  octets
ATTRIBUTE   Framed-AppleTalk-Link   37  integer
ATTRIBUTE   Framed-AppleTalk-Network 38 integer
ATTRIBUTE   Framed-AppleTalk-Zone   39  string
"""

REALM_SEPARATOR = '@'


class RADIUSBackend(object):
    """
    Standard RADIUS authentication backend for Django. Uses the server details
    specified in settings.py (RADIUS_SERVER, RADIUS_PORT and RADIUS_SECRET).
    """
    supports_anonymous_user = False
    supports_object_permissions = False

    def __init__(self, settings=None):
        if settings is None:
            settings = global_settings
        self.settings = settings

    def _get_dictionary(self):
        """
        Get the pyrad Dictionary object which will contain our RADIUS user's
        attributes. Fakes a file-like object using StringIO.
        """
        return Dictionary(StringIO(DICTIONARY))

    def _get_auth_packet(self, username, password, client):
        """
        Get the pyrad authentication packet for the username/password and the
        given pyrad client.
        """
        pkt = client.CreateAuthPacket(code=AccessRequest,
                                      User_Name=username)
        pkt["User-Password"] = pkt.PwCrypt(password)
        pkt["NAS-Identifier"] = 'django-radius'
        for key, val in list(getattr(self.settings, 'RADIUS_ATTRIBUTES', {}).items()):
            pkt[key] = val
        return pkt

    def _get_client(self, server):
        """
        Get the pyrad client for a given server. RADIUS server is described by
        a 3-tuple: (<hostname>, <port>, <secret>).
        """
        return Client(
            server=server[0],
            authport=server[1],
            secret=server[2],
            dict=self._get_dictionary(),
        )

    def _get_server_from_settings(self):
        """
        Get the RADIUS server details from the settings file.
        """
        return (
            self.settings.RADIUS_SERVER,
            int(self.settings.RADIUS_PORT),
            self.settings.RADIUS_SECRET.encode('utf-8'),
        )

    def _perform_radius_auth(self, client, packet):
        """
        Perform the actual radius authentication by passing the given packet
        to the server which `client` is bound to.
        Returns a tuple (list of groups, is_staff, is_superuser) or None depending on whether the user is authenticated
        successfully.
        """
        try:
            reply = client.SendPacket(packet)
        except Timeout as e:
            logger.error("RADIUS timeout occurred contacting %s:%s" % (
                client.server, client.authport))
            return None
        except Exception as e:
            logger.error("RADIUS error: %s" % e)
            return None

        if reply.code == AccessReject:
            logger.warning("RADIUS access rejected for user '%s'" % (
                packet['User-Name']))
            return None
        elif reply.code != AccessAccept:
            logger.error("RADIUS access error for user '%s' (code %s)" % (
                packet['User-Name'], reply.code))
            return None

        logger.info("RADIUS access granted for user '%s'" % (
            packet['User-Name']))

        if "Class" not in reply.keys():
            return [], False, False

        groups = []
        is_staff = False
        is_superuser = False

        app_class_prefix = getattr(self.settings, 'RADIUS_CLASS_APP_PREFIX', '')
        group_class_prefix = app_class_prefix + "group="
        role_class_prefix = app_class_prefix + "role="

        for cl in reply['Class']:
            cl = cl.decode("utf-8")
            if cl.lower().find(group_class_prefix) == 0:
                groups.append(cl[len(group_class_prefix):])
            elif cl.lower().find(role_class_prefix) == 0:
                role = cl[len(role_class_prefix):]
                if role == "staff":
                    is_staff = True
                elif role == "superuser":
                    is_superuser = True
                else:
                    logger.warning("RADIUS Attribute Class contains unknown role '%s'. Only roles 'staff' and 'superuser' are allowed" % cl)
        return groups, is_staff, is_superuser

    def _radius_auth(self, server, username, password):
        """
        Authenticate the given username/password against the RADIUS server
        described by `server`.
        """
        client = self._get_client(server)
        packet = self._get_auth_packet(username, password, client)
        return self._perform_radius_auth(client, packet)

    def get_django_user(self, username, password=None, groups=None, is_staff=False, is_superuser=False):
        """
        Get the Django user with the given username, or create one if it
        doesn't already exist. If `password` is given, then set the user's
        password to that (regardless of whether the user was created or not).
        """
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            user = User(username=username)

        # if RADIUS_REMOTE_ROLES is not set, configure it to the default value
        # of versions <= 1.4.0
        if not hasattr(self.settings, "RADIUS_REMOTE_ROLES"):
            self.settings.RADIUS_REMOTE_ROLES = True

        if self.settings.RADIUS_REMOTE_ROLES:
            user.is_staff = is_staff
            user.is_superuser = is_superuser
        if password is not None:
            user.set_password(password)

        user.save()
        user.groups.set(groups)
        return user

    def get_user_groups(self, group_names):
        groups = Group.objects.filter(name__in=group_names)
        if len(groups) != len(group_names):
            local_group_names = [g.name for g in groups]
            logger.warning("RADIUS reply contains %d user groups (%s), but only %d (%s) found" % (
                len(group_names), ", ".join(group_names), len(groups), ", ".join(local_group_names)))
        return groups

    def authenticate(self, request, username=None, password=None):
        """
        Check credentials against RADIUS server and return a User object or
        None.
        """

        server = self._get_server_from_settings()
        result = self._radius_auth(server, username, password)

        if result:
            group_names, is_staff, is_superuser = result
            groups = self.get_user_groups(group_names)
            return self.get_django_user(username, password, groups, is_staff, is_superuser)

        return None

    def get_user(self, user_id):
        """
        Get the user with the ID of `user_id`. Authentication backends must
        implement this method.
        """
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None


class RADIUSRealmBackend(RADIUSBackend):
    """
    Advanced realm-based RADIUS backend. Authenticates users with a username,
    password and a realm (simply a unique string). The server to authenticate
    with is defined by the result of calling get_server(realm) on an instance
    of this class.

    By default, this class uses the RADIUS server specified in the settings
    file, regardless of the realm. Subclasses should override the `get_server`
    method to provide their own logic. The method should return a 3-tuple:
    (<hostname>, <port>, <secret>).
    """
    def get_server(self, realm):
        """
        Get the details of the RADIUS server to authenticate users of the given
        realm.

        Returns a 3-tuple (<hostname>, <port>, <secret>). Base implementation
        always returns the RADIUS server specified in the main settings file,
        and should be overridden.
        """
        return self._get_server_from_settings()

    def construct_full_username(self, username, realm):
        """
        Construct a unique username for a user, given their normal username and
        realm. This is to avoid conflicts in the Django auth app, as usernames
        must be unique.

        By default, returns a string in the format <username>@<realm>.
        """
        return '%s@%s' % (username, realm)

    def authenticate(self, request, username=None, password=None, realm=None):
        """
        Check credentials against the RADIUS server identified by `realm` and
        return a User object or None. If no argument is supplied, Django will
        skip this backend and try the next one (as a TypeError will be raised
        and caught).
        """

        server = self.get_server(realm)

        if not server:
            return None

        result = self._radius_auth(server, username, password)

        if result:
            full_username = self.construct_full_username(username, realm)
            group_names, is_staff, is_superuser = result
            groups = self.get_user_groups(group_names)
            return self.get_django_user(full_username, password, groups, is_staff, is_superuser)

        return None



import hashlib
import logging

from django.utils.encoding import smart_str
from oauth2_provider.contrib.rest_framework import OAuth2Authentication
from oauth2_provider.oauth2_backends import OAuthLibCore as _OAuthLibCore
from rest_framework.exceptions import UnsupportedMediaType

from ansible_base.lib.utils.hashing import hash_string

logger = logging.getLogger('ansible_base.oauth2_provider.authentication')


class OAuthLibCore(_OAuthLibCore):
    def extract_body(self, request):
        try:
            return request.POST.items()
        except UnsupportedMediaType:
            return ()


class LoggedOAuth2Authentication(OAuth2Authentication):
    def authenticate(self, request):
        # sha256 the bearer token. We store the hash in the database
        # and this gives us a place to hash the incoming token for comparison
        did_hash_token = False
        bearer_token = request.META.get('HTTP_AUTHORIZATION')
        if bearer_token and bearer_token.lower().startswith('bearer '):
            token_component = bearer_token.split(' ', 1)[1]
            hashed = hash_string(token_component, hasher=hashlib.sha256, algo="sha256")
            did_hash_token = True
            request.META['HTTP_AUTHORIZATION'] = f"Bearer {hashed}"

        # We don't /really/ want to modify the request, so after we're done authing,
        # revert what we did above.
        try:
            ret = super().authenticate(request)
        finally:
            if did_hash_token:
                request.META['HTTP_AUTHORIZATION'] = bearer_token

        if ret:
            user, token = ret
            username = user.username if user else '<none>'
            logger.info(
                smart_str(u"User {} performed a {} to {} through the API using OAuth 2 token {}.".format(username, request.method, request.path, token.pk))
            )
            # TODO: check oauth_scopes when we have RBAC in Gateway
            setattr(user, 'oauth_scopes', [x for x in token.scope.split() if x])
        return ret



# Django
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from ansible_base.oauth2_provider.serializers import OAuth2TokenSerializer

User = get_user_model()


class Command(BaseCommand):
    """Command that creates an OAuth2 token for a certain user. Returns the value of created token."""

    help = 'Creates an OAuth2 token for a user.'

    def add_arguments(self, parser):
        parser.add_argument('--user', dest='user', type=str)

    def handle(self, *args, **options):
        if not options['user']:
            raise CommandError('Username not supplied. Usage: create_oauth2_token --user=username.')
        try:
            user = User.objects.get(username=options['user'])
        except ObjectDoesNotExist:
            raise CommandError('The user does not exist.')
        config = {'user': user, 'scope': 'write'}
        serializer_obj = OAuth2TokenSerializer()

        class FakeRequest(object):
            def __init__(self):
                self.user = user

        serializer_obj.context['request'] = FakeRequest()
        serializer_obj.create(config)
        self.stdout.write(serializer_obj.unencrypted_token)



from django.contrib import admin

from ansible_base.authentication.models import Authenticator, AuthenticatorMap, AuthenticatorUser

admin.site.register(Authenticator)
admin.site.register(AuthenticatorMap)
admin.site.register(AuthenticatorUser)



import django.apps
from django.core.checks import Error, register
from django.db import models


@register()
def check_charfield_has_max_length(app_configs, **kwargs):
    errors = []
    for model in django.apps.apps.get_models():
        for field in model._meta.fields:
            if isinstance(field, models.CharField) and field.max_length is None:
                errors.append(
                    Error(
                        'CharField must have a max_length',
                        hint=f"Add max_length parameter for field '{field.name}' in {model.__name__}",
                        id='ansible_base.E001',
                    )
                )
    return errors



# Generated by Django 4.2.16 on 2024-11-21 11:14

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        migrations.swappable_dependency(settings.ANSIBLE_BASE_TEAM_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
        ('dab_rbac', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='objectrole',
            name='provides_teams',
            field=models.ManyToManyField(editable=False, help_text='Users who have this role obtain member access to these teams, and inherit all their permissions.', related_name='member_roles', to=settings.ANSIBLE_BASE_TEAM_MODEL),
        ),
        migrations.AlterField(
            model_name='objectrole',
            name='role_definition',
            field=models.ForeignKey(help_text='The role definition which defines what permissions this object role grants.', on_delete=django.db.models.deletion.CASCADE, related_name='object_roles', to='dab_rbac.roledefinition'),
        ),
        migrations.AlterField(
            model_name='objectrole',
            name='teams',
            field=models.ManyToManyField(help_text='Teams or groups who have access to the permissions defined by this object role.', related_name='has_roles', through='dab_rbac.RoleTeamAssignment', to=settings.ANSIBLE_BASE_TEAM_MODEL),
        ),
        migrations.AlterField(
            model_name='objectrole',
            name='users',
            field=models.ManyToManyField(help_text='Users who have access to the permissions defined by this object role.', related_name='has_roles', through='dab_rbac.RoleUserAssignment', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='content_type',
            field=models.ForeignKey(default=None, help_text='Thw type of resource this can apply to; only used for validation and user assistance.', null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='roleevaluation',
            name='codename',
            field=models.TextField(help_text='The name of the permission, giving the action and the model, from the Django Permission model.'),
        ),
        migrations.AlterField(
            model_name='roleevaluation',
            name='role',
            field=models.ForeignKey(help_text='The object role that grants this form of permission.', on_delete=django.db.models.deletion.CASCADE, related_name='permission_partials', to='dab_rbac.objectrole'),
        ),
        migrations.AlterField(
            model_name='roleevaluationuuid',
            name='codename',
            field=models.TextField(help_text='The name of the permission, giving the action and the model, from the Django Permission model.'),
        ),
        migrations.AlterField(
            model_name='roleevaluationuuid',
            name='role',
            field=models.ForeignKey(help_text='The object role that grants this form of permission.', on_delete=django.db.models.deletion.CASCADE, related_name='permission_partials_uuid', to='dab_rbac.objectrole'),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='object_id',
            field=models.TextField(blank=True, help_text='The primary key of the object this assignment applies to; null value indicates system-wide assignment.', null=True),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='role_definition',
            field=models.ForeignKey(help_text='The role definition which defines permissions conveyed by this assignment.', on_delete=django.db.models.deletion.CASCADE, related_name='team_assignments', to='dab_rbac.roledefinition'),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='object_id',
            field=models.TextField(blank=True, help_text='The primary key of the object this assignment applies to; null value indicates system-wide assignment.', null=True),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='role_definition',
            field=models.ForeignKey(help_text='The role definition which defines permissions conveyed by this assignment.', on_delete=django.db.models.deletion.CASCADE, related_name='user_assignments', to='dab_rbac.roledefinition'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='content_type',
            field=models.ForeignKey(default=None, help_text='The type of resource this can apply to; only used for validation and user assistance.', null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
    ]



import oauth2_provider.models as oauth2_models
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from oauth2_provider.generators import generate_client_secret

from ansible_base.lib.abstract_models.common import NamedCommonModel
from ansible_base.lib.utils.models import prevent_search
from ansible_base.lib.utils.response import get_relative_url

activitystream = object
if 'ansible_base.activitystream' in settings.INSTALLED_APPS:
    from ansible_base.activitystream.models import AuditableModel

    activitystream = AuditableModel


class OAuth2Application(NamedCommonModel, oauth2_models.AbstractApplication, activitystream):
    router_basename = 'application'
    ignore_relations = ['oauth2idtoken', 'grant', 'oauth2refreshtoken']
    # We do NOT add client_secret to encrypted_fields because it is hashed by Django OAuth Toolkit
    # and it would end up hashing the encrypted value.

    class Meta(oauth2_models.AbstractAccessToken.Meta):
        verbose_name = _('application')
        unique_together = (("name", "organization"),)
        ordering = ('organization', 'name')
        swappable = "OAUTH2_PROVIDER_APPLICATION_MODEL"

    CLIENT_TYPES = (
        ("confidential", _("Confidential")),
        ("public", _("Public")),
    )

    GRANT_TYPES = (
        ("authorization-code", _("Authorization code")),
        ("password", _("Resource owner password-based")),
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="applications",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        help_text=_("The user who created the application."),
    )

    description = models.TextField(
        default='',
        blank=True,
        help_text=_("A description of this application."),
    )
    organization = models.ForeignKey(
        getattr(settings, 'ANSIBLE_BASE_ORGANIZATION_MODEL'),
        related_name='applications',
        help_text=_('Organization containing this application.'),
        on_delete=models.CASCADE,
        null=True,
    )

    app_url = models.URLField(
        default=None,
        blank=True,
        null=True,
        help_text=_("The URL of this application."),
    )

    # It would be nice to just use our usual encrypted_fields flow here
    # until DOT makes a release with https://github.com/jazzband/django-oauth-toolkit/pull/1311
    # there is no way to disable its expectation of using its own hashing
    # (which is Django's make_password/check_password).
    # So we use their field here.
    # Previous versions of DOT didn't hash the field at all and AWX pins
    # to <2.0.0 so AWX used the AWX encryption with no issue.
    # We still override it here so that we can prevent_search() on it.
    client_secret = prevent_search(
        oauth2_models.ClientSecretField(
            max_length=255,
            blank=True,
            default=generate_client_secret,
            db_index=True,
            help_text=_("Hashed on Save. Copy it now if this is a new secret."),
        )
    )

    client_type = models.CharField(
        max_length=32, choices=CLIENT_TYPES, help_text=_('Set to Public or Confidential depending on how secure the client device is.')
    )
    skip_authorization = models.BooleanField(default=False, help_text=_('Set True to skip authorization step for completely trusted applications.'))
    authorization_grant_type = models.CharField(
        max_length=32, choices=GRANT_TYPES, help_text=_('The Grant type the user must use for acquire tokens for this application.')
    )
    updated = None  # Tracked in CommonModel with 'modified', no need for this

    def get_absolute_url(self):
        # This is kind of annoying. This method lives on the superclass and we check for it in CommonModel.
        # But better would be to not have this method and let the CommonModel logic fall back to the "right" way of finding this.
        return get_relative_url(f'{self.router_basename}-detail', kwargs={'pk': self.pk})



# Django
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from ansible_base.oauth2_provider.models import OAuth2AccessToken, OAuth2RefreshToken

User = get_user_model()


def revoke_tokens(token_list):
    for token in token_list:
        token.revoke()
        print('revoked {} {}'.format(token.__class__.__name__, token.token))


class Command(BaseCommand):
    """Command that revokes OAuth2 access tokens."""

    help = 'Revokes OAuth2 access tokens.  Use --all to revoke access and refresh tokens.'

    def add_arguments(self, parser):
        parser.add_argument('--user', dest='user', type=str, help='revoke OAuth2 tokens for a specific username')
        parser.add_argument('--all', dest='all', action='store_true', help='revoke OAuth2 access tokens and refresh tokens')

    def handle(self, *args, **options):
        if not options['user']:
            if options['all']:
                revoke_tokens(OAuth2RefreshToken.objects.filter(revoked=None))
            revoke_tokens(OAuth2AccessToken.objects.all())
        else:
            try:
                user = User.objects.get(username=options['user'])
            except ObjectDoesNotExist:
                raise CommandError('A user with that username does not exist.')
            if options['all']:
                revoke_tokens(OAuth2RefreshToken.objects.filter(revoked=None).filter(user=user))
            revoke_tokens(OAuth2AccessToken.objects.filter(user=user))



import logging

from rest_framework import serializers

logger = logging.getLogger('ansible_base.lib.serializers.mixins')


# Derived from: https://github.com/encode/django-rest-framework/discussions/8606
class ImmutableFieldsMixin(serializers.ModelSerializer):
    # Mixin enabling the usage of Meta.immutable_fields for setting fields read_only after object creation.

    # Currently, using this without issues requires outside considerations:
    #     1. overrides to get_serializer for the related viewsets,
    #        since by default, rest_framework's SimpleMetadata class does not try to provide initialize a serializer
    #        with an instance value on elements with a primary key field.

    #        See ansible_base.authentication.views.AuthenticatorViewSet for an example.
    #    2. The generated OpenAPI spec will treat immutable fields as valid parameters on PUT and PATCH endpoints

    def get_extra_kwargs(self):
        kwargs = super().get_extra_kwargs()
        immutable_fields = getattr(self.Meta, "immutable_fields", [])

        # Make field read_only if instance already exists
        for field in immutable_fields:
            kwargs.setdefault(field, {})
            kwargs[field]["read_only"] = bool(self.instance)

        return kwargs



import oauth2_provider.models as oauth2_models
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from ansible_base.lib.abstract_models.common import CommonModel

activitystream = object
if 'ansible_base.activitystream' in settings.INSTALLED_APPS:
    from ansible_base.activitystream.models import AuditableModel

    activitystream = AuditableModel


class OAuth2IDToken(CommonModel, oauth2_models.AbstractIDToken, activitystream):
    class Meta(oauth2_models.AbstractIDToken.Meta):
        verbose_name = _('id token')
        swappable = "OAUTH2_PROVIDER_ID_TOKEN_MODEL"

    updated = None  # Tracked in CommonModel with 'modified', no need for this



import logging

from django.utils.translation import gettext_lazy as _
from social_core.backends.google import GoogleOAuth2

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.social_auth import SocialAuthMixin, SocialAuthValidateCallbackMixin
from ansible_base.lib.serializers.fields import BooleanField, CharField, ChoiceField, ListField, URLField

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.google_oauth2')


class GoogleOAuth2Configuration(BaseAuthenticatorConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/google.html#google-oauth2"

    #################################
    # Minimal params
    #################################

    KEY = CharField(
        help_text=_("The OAuth2 key from your web application."),
        allow_null=False,
        ui_field_label=_('Google OAuth2 Key'),
    )

    SECRET = CharField(
        help_text=_("The OAuth2 secret from your web application."),
        allow_null=True,
        ui_field_label=_('Google OAuth2 Secret'),
    )

    #################################
    # Additional params
    #################################

    CALLBACK_URL = URLField(
        help_text=_(
            'Provide this URL as the callback URL for your application as part of your registration process. Refer to the documentation for more detail.'
        ),
        required=False,
        allow_null=True,
        ui_field_label=_("Callback URL"),
    )

    ACCESS_TOKEN_URL = URLField(
        help_text=_("The URL to obtain an access token from the OIDC provider."),
        required=False,
        allow_null=True,
        ui_field_label=_("Access Token URL"),
    )

    ACCESS_TOKEN_METHOD = ChoiceField(
        help_text=_("The HTTP method to be used when requesting an access token. Typically 'POST' or 'GET'."),
        default="POST",
        allow_null=True,
        required=False,
        choices=['GET', 'PUT', 'POST', 'PATCH', 'DELETE'],
        ui_field_label=_("Access Token Method"),
    )

    AUTHORIZATION_URL = URLField(
        help_text=_("The URL to redirect the user for Google OAuth2  provider authorization."),
        required=False,
        allow_null=True,
        ui_field_label=_("Authorization URL"),
    )

    REDIRECT_STATE = BooleanField(
        help_text=_("Enable or disable state parameter in the redirect URI. Recommended to be True for preventing CSRF attacks."),
        default=False,
        allow_null=True,
        ui_field_label=_("Redirect State"),
    )

    REVOKE_TOKEN_METHOD = ChoiceField(
        help_text=_("The HTTP method to be used when revoking an access token. Typically 'POST' or 'GET'."),
        default="GET",
        allow_null=True,
        choices=['GET', 'PUT', 'POST', 'PATCH', 'DELETE'],
        ui_field_label=_("Revoke Token Method"),
    )

    REVOKE_TOKEN_URL = URLField(
        help_text=_("The URL to revoke tokens. Used in the token revocation flow."),
        required=False,
        allow_null=True,
        ui_field_label=_("Revoke Token URL"),
    )

    SCOPE = ListField(
        help_text=_('The authorization scope for users. Defaults to ["openid", "email", "profile"].'),
        required=False,
        allow_null=False,
        ui_field_label=_('List of OAuth2 Scope(s)'),
        default=["openid", "email", "profile"],
    )


class AuthenticatorPlugin(SocialAuthMixin, SocialAuthValidateCallbackMixin, GoogleOAuth2, AbstractAuthenticatorPlugin):
    configuration_class = GoogleOAuth2Configuration
    type = "google_oauth2"
    logger = logger
    category = "sso"
    configuration_encrypted_fields = ['SECRET']



from django.contrib import admin


class ReadOnlyAdmin(admin.ModelAdmin):
    """
    A ModelAdmin that provides read-only access to the model.
    """

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False



# Generated by Django 4.2.8 on 2024-03-18 14:59

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created')),
                ('object_id', models.TextField(blank=True, null=True)),
                ('operation', models.CharField(choices=[('create', 'Entity created'), ('update', 'Entity updated'), ('delete', 'Entity deleted'), ('associate', 'Entity was associated with another entity'), ('disassociate', 'Entity was disassociated with another entity')], max_length=12)),
                ('changes', models.JSONField(blank=True, null=True)),
                ('related_object_id', models.TextField(blank=True, null=True)),
                ('related_field_name', models.CharField(blank=True, max_length=64, null=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.contenttype')),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('related_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='related_content_type', to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name_plural': 'Entries',
            },
        ),
    ]



import logging
from dataclasses import dataclass
from types import SimpleNamespace

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from ansible_base.authentication.authenticator_plugins._radiusauth import RADIUSBackend as _RADIUSBackend
from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.utils.authentication import get_or_create_authenticator_user
from ansible_base.authentication.utils.claims import update_user_claims
from ansible_base.lib.serializers.fields import CharField, IntegerField

logger = logging.getLogger(__name__)

User = get_user_model()


@dataclass
class RADIUSUser:
    username: str
    groups: list[str]
    is_staff: bool
    is_superuser: bool


class RADIUSBackend(_RADIUSBackend):
    def get_django_user(self, username, password=None, groups=None, is_staff=False, is_superuser=False):
        return RADIUSUser(
            username=username,
            groups=groups or [],
            is_staff=is_staff,
            is_superuser=is_superuser,
        )

    def get_user_groups(self, group_names):
        return group_names


class RADIUSConfiguration(BaseAuthenticatorConfiguration):
    SERVER = CharField(
        label=_("RADIUS Server"),
        help_text="Hostname of RADIUS server.",
        ui_field_label=_('Hostname of RADIUS Server'),
        required=True,
    )

    PORT = IntegerField(
        min_value=1,
        max_value=65535,
        default=1812,
        label=_("RADIUS Port"),
        help_text=_("Port number of RADIUS server."),
        ui_field_label=_("Port number of RADIUS Server"),
    )

    SECRET = CharField(
        label=_("RADIUS Secret"),
        help_text=_("Shared secret for authenticating to RADIUS server."),
        ui_field_label=_("Shared secret for authenticating to RADIUS server."),
        required=True,
    )


class AuthenticatorPlugin(AbstractAuthenticatorPlugin):
    configuration_class = RADIUSConfiguration
    type = "RADIUS"
    category = "password"
    configuration_encrypted_fields = ["SECRET"]

    def __init__(self, database_instance=None, *args, **kwargs):
        super().__init__(database_instance, *args, **kwargs)
        self.set_logger(logger)

    def authenticate(self, request, username=None, password=None, **kwargs):
        if not username or not password:
            return None

        settings = SimpleNamespace(
            RADIUS_SERVER=self.settings["SERVER"],
            RADIUS_PORT=self.settings["PORT"],
            RADIUS_SECRET=self.settings["SECRET"],
        )
        backend = RADIUSBackend(settings)
        radius_user = backend.authenticate(request, username, password)
        if radius_user is None:
            return None

        user, _authenticator_user, _is_created = get_or_create_authenticator_user(
            username,
            authenticator=self.database_instance,
            user_details={},
            extra_data={"username": username},
        )
        return update_user_claims(user, self.database_instance, radius_user.groups)



from rest_framework.response import Response

from ansible_base.authentication.utils.trigger_definition import TRIGGER_DEFINITION
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView


class TriggerDefinitionView(AnsibleBaseDjangoAppApiView):
    def get(self, request, format=None):
        return Response(TRIGGER_DEFINITION)



from typing import Type

from rest_framework.schemas.generators import EndpointEnumerator
from rest_framework.views import APIView


def get_api_view_functions(urlpatterns=None) -> set[Type[APIView]]:
    """
    Extract view classes from a urlpatterns list using the show_urls helper functions

    :param urlpatterns: django urlpatterns list
    :return: set of all view classes used by the urlpatterns list
    """
    views = set()

    enumerator = EndpointEnumerator()
    # Get all active APIViews from urlconf
    endpoints = enumerator.get_api_endpoints(patterns=urlpatterns)
    for _, _, func in endpoints:
        # ApiView.as_view() breadcrumb
        if hasattr(func, 'cls'):
            views.add(func.cls)

    return views



# Generated by Django 4.2.11 on 2024-09-10 14:32

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0013_alter_authenticator_order'),
    ]

    operations = [
        migrations.AddField(
            model_name='authenticator',
            name='auto_migrate_users_to',
            field=models.ForeignKey(help_text='Automatically move users from this authenticator to the target authenticator when a matching user logs in via the target authenticator. For this to work, the field used for the user ID on both authenticators needs to have the same value. This should only be used when migrating users between two authentication mechanisms that share the same user database (such as when both IDPs share the same LDAP user directory).', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='auto_migrate_users_from', to='dab_authentication.authenticator'),
        ),
    ]



import logging
import uuid

from ansible_base.lib.logging import thread_local
from ansible_base.lib.utils.collection import first_matching


class RequestIdFilter(logging.Filter):
    def filter(self, record):
        """
        This "filter" is used to add the request id to the log record.
        It will always return True, so that the message is always logged,
        even if there is no request id.
        """

        # Always a default so we can use it in the logging formatter.
        record.request_id = ""

        request = getattr(thread_local, "request", None)

        if request is None:
            # request never got added to the thread local, so we can't add the request id to the log.
            # But we still want to log the message.
            return True

        headers = request.META.keys()
        request_id_header = first_matching(lambda x: x.lower().replace('-', '_') == "http_x_request_id", headers, default=None)
        if request_id_header is None:
            # We have a request, but no request id header. Still want to log the message.
            return True
        request_id = request.META.get(request_id_header)
        try:
            uuid.UUID(request_id)
        except ValueError:
            # Invalid request id, still want to log the message, but not with the invalid request id.
            pass
        else:
            record.request_id = request_id
        return True



from django.db.models import JSONField
from django.db.models.fields.related import ForeignObjectRel
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import ParseError, PermissionDenied

from ansible_base.lib.utils.models import get_all_field_names


def get_fields_from_path(model, path, treat_jsonfield_as_text=True):
    """
    Given a Django ORM lookup path (possibly over multiple models)
    Returns the fields in the line, and also the revised lookup path
    ex., given
        model=Organization
        path='project__timeout'
    returns tuple of fields traversed as well and a corrected path,
    for special cases we do substitutions
        ([<IntegerField for timeout>], 'project__timeout')
    """
    # Store of all the fields used to detect repeats
    field_list = []
    new_parts = []
    is_json_field = False

    for name in path.split('__'):
        if is_json_field:
            new_parts.append(name)
            continue

        if model is None:
            raise ParseError(_('No related model for field %(model_name)s.') % {'model_name': name})
        # TODO: Do we want to keep these AWX specific items here?
        # HACK: Make project and inventory source filtering by old field names work for backwards compatibility.
        if model._meta.object_name in ('Project', 'InventorySource'):
            name = {'current_update': 'current_job', 'last_update': 'last_job', 'last_update_failed': 'last_job_failed', 'last_updated': 'last_job_run'}.get(
                name, name
            )

        if name == 'type' and 'polymorphic_ctype' in get_all_field_names(model):
            name = 'polymorphic_ctype'
            new_parts.append('polymorphic_ctype__model')
        else:
            new_parts.append(name)

        if name in getattr(model, 'PASSWORD_FIELDS', ()) or name in getattr(model, 'encrypted_fields', ()):
            raise PermissionDenied(_('Filtering on field %(name)s is not allowed.') % {'name': name})
        elif name == 'pk':
            field = model._meta.pk
        else:
            name_alt = name.replace("_", "")
            if name_alt in model._meta.fields_map.keys():
                field = model._meta.fields_map[name_alt]
                new_parts.pop()
                new_parts.append(name_alt)
            else:
                field = model._meta.get_field(name)
            if isinstance(field, ForeignObjectRel) and getattr(field.field, '__prevent_search__', False):
                raise PermissionDenied(_('Filtering on %(field_name)s is not allowed.') % {"field_name": name})
            elif getattr(field, '__prevent_search__', False):
                raise PermissionDenied(_('Filtering on %(field_name)s is not allowed.') % {"field_name": name})
            elif isinstance(field, JSONField) and not treat_jsonfield_as_text:
                is_json_field = True
        if field in field_list:
            # Field traversed twice, could create infinite JOINs, DoS-ing the service
            raise ParseError(_('Loops not allowed in filters, detected on field %(field_name)s.') % {'field_name': field.name})
        field_list.append(field)
        model = getattr(field, 'related_model', None)

    return field_list, '__'.join(new_parts)


def get_field_from_path(model, path):
    """
    Given a Django ORM lookup path (possibly over multiple models)
    Returns the last field in the line, and the revised lookup path
    ex.
        (<IntegerField for timeout>, 'project__timeout')
    """
    field_list, new_path = get_fields_from_path(model, path)
    return (field_list[-1], new_path)



import logging

from django.contrib.sessions.backends.cached_db import SessionStore as CachedDBSessionStore
from django.utils.translation import gettext_lazy as _

from ansible_base.lib.utils.settings import get_setting

DEFAULT_SESSION_TIMEOUT = 30 * 60
logger = logging.getLogger('ansible_base.lib.sessions.stores.cached_dynamic_timeout')


class SessionStore(CachedDBSessionStore):
    cache_key_prefix = 'ansible_base.lib.sessions.stores.cached_dynamic_timeout'

    def get_session_cookie_age(self):
        timeout = get_setting('SESSION_COOKIE_AGE', DEFAULT_SESSION_TIMEOUT)
        if not isinstance(timeout, int):
            logger.error(
                _('SESSION_COOKIE_AGE was set to %(timeout)s which is an invalid int, defaulting to %(default)s')
                % {'timeout': timeout, 'default': DEFAULT_SESSION_TIMEOUT}
            )
            timeout = DEFAULT_SESSION_TIMEOUT
        return timeout



# Generated by Django 4.2.6 on 2024-02-14 16:31

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0002_alter_authenticator_users'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticatormap',
            name='authenticator',
            field=models.ForeignKey(help_text='The authenticator this mapping belongs to', on_delete=django.db.models.deletion.CASCADE, related_name='authenticator_maps', to='dab_authentication.authenticator'),
        ),
    ]



from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from itertools import chain
from typing import TYPE_CHECKING, Optional

from crum import get_current_user
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _
from inflection import underscore

from ansible_base.lib.utils.create_system_user import create_system_user, get_system_username
from ansible_base.lib.utils.encryption import ENCRYPTED_STRING
from ansible_base.lib.utils.string import make_json_safe

logger = logging.getLogger('ansible_base.lib.utils.models')

# Handle type hints of AbstractUser during linting
if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser


def get_all_field_names(model, concrete_only=False, include_attnames=True):
    # Implements compatibility with _meta.get_all_field_names
    # See: https://docs.djangoproject.com/en/1.11/ref/models/meta/#migrating-from-the-old-api
    return list(
        set(
            chain.from_iterable(
                (field.name, field.attname) if include_attnames and hasattr(field, 'attname') else (field.name,)
                for field in model._meta.get_fields()
                # For complete backwards compatibility, you may want to exclude
                # GenericForeignKey from the results.
                if not (field.many_to_one and field.related_model is None) and not (concrete_only and not field.concrete)
            )
        )
    )


def get_type_for_model(model):
    """
    Return type name for a given model class.
    """
    opts = model._meta.concrete_model._meta
    return underscore(opts.object_name)


def is_add_perm(codename: str) -> bool:
    """Utility method to say whether codename represents adding permission.

    This type of permission requires special treatment in several places
    in DAB RBAC because the permission defers to the parent object.
    Although the function is trivial, this standardizes the criteria.
    """
    use_codename = codename
    if codename.count('.') == 1:
        _, use_codename = codename.split('.')
    return use_codename.startswith('add_')


def prevent_search(relation):
    """
    Used to mark a model field or relation as "restricted from filtering"
    e.g.,

    class AuthToken(BaseModel):
        user = prevent_search(models.ForeignKey(...))
        sensitive_data = prevent_search(models.CharField(...))

    The flag set by this function is used by
    `ansible_base.rest_filters.rest_framework.field_lookup_backend.FieldLookupBackend` to block fields and relations that
    should not be searchable/filterable via search query params
    """
    setattr(relation, '__prevent_search__', True)
    return relation


def user_summary_fields(user):
    sf = {}
    for field_name in ('id', 'username', 'first_name', 'last_name'):
        sf[field_name] = getattr(user, field_name)
    return sf


def is_system_user(user: Optional[models.Model]) -> bool:
    """
    Takes a model and returns a boolean if its a user whose username is the same as the SYSTEM_USERNAME setting
    """
    from django.contrib.auth.models import AbstractUser

    system_username = get_system_username()[0]
    if user is None or not isinstance(user, AbstractUser) or system_username is None:
        # If we didn't get anything or that thing isn't an AbstractUser or system_username is not set set than what we have can't be the system user
        return False
    return user.username == system_username


class NotARealException(Exception):
    pass


def get_system_user() -> Optional[AbstractUser]:

    from ansible_base.lib.abstract_models.user import AbstractDABUser

    system_username, setting_name = get_system_username()
    user_model = get_user_model()

    # If we use subclass of AbstractDABUser ensure we use manager for unfiltered queryset
    user_manager = user_model.all_objects if issubclass(user_model, AbstractDABUser) else user_model.objects

    system_user = user_manager.filter(username=system_username).first()
    # We are using a global variable to try and track if this thread has already spit out the message, if so ignore
    if system_username is not None and system_user is None:
        logger.error(
            _(
                "{setting_name} is set to {system_username} but no user with that username exists.".format(
                    setting_name=setting_name, system_username=system_username
                )
            )
        )
        caught_exception = NotARealException
        if 'ansible_base.resource_registry' in settings.INSTALLED_APPS:
            from ansible_base.resource_registry.models import ResourceType

            caught_exception = ResourceType.DoesNotExist
            # If resource registry is installed we hit issues here during test tear downs
            # For an unidentified reason, during teardown, the tests are calling the post_migration signals from resource_registry
            # These eventually call get_or_create on models which then try and call current_or_system_user which eventually leads here
            # But the system is in a weird state here because its being torn down, so the creation of system_user fails
        try:
            system_user = create_system_user(user_model=get_user_model())
        except caught_exception:
            system_user = None

    return system_user


def current_user_or_system_user() -> Optional[AbstractUser]:
    """
    Attempt to get the current user. If there is none or it is anonymous,
    try to return the system user instead.
    """
    user = get_current_user()
    if user is None or user.is_anonymous:
        user = get_system_user()
    return user


def is_encrypted_field(model, field_name):
    from django.contrib.auth.models import AbstractUser

    if model is None:
        return False

    if issubclass(model, AbstractUser) and field_name == 'password':
        return True

    # This throws FieldDoesNotExist if the field does not exist, which is reasonable here, so we don't catch it
    field = model._meta.get_field(field_name)
    if getattr(field, '__prevent_search__', False):
        return True

    return field_name in getattr(model, 'encrypted_fields', [])


@dataclass
class ModelDiff:
    added_fields: dict
    removed_fields: dict
    changed_fields: dict

    def __bool__(self):
        return bool(self.added_fields or self.removed_fields or self.changed_fields)

    @property
    def has_changes(self):
        return bool(self)

    dict = asdict


def diff(
    old,
    new,
    require_type_match=True,
    json_safe=True,
    include_m2m=False,
    exclude_fields=[],
    limit_fields=[],
    sanitize_encrypted=True,
    all_values_as_strings=False,
):
    """
    Diff two instances of models (which do not have to be the same type of model
    if given require_type_match=False).

    This function is used in particular by the activitystream application where
    the changes returned by this function are stored as models change.

    :param old: The old instance for comparison
    :param new: The new instance for comparison
    :param require_type_match: If True, old and new must be of the same type of
        model. (default: True)
    :param json_safe: If True, the diff will be made JSON-safe by converting
        all non-JSON-safe values to strings using Django's smart_str function.
        (default: True)
    :param include_m2m: If True, include many-to-many fields in the diff.
        Otherwise, they are ignored. (default: False)
    :param exclude_fields: A list of field names to exclude from the diff.
        (default: [])
    :param limit_fields: A list of field names to limit the diff to. This can be
        useful, for example, when update_fields is passed to a model's save
        method and you only want to diff the fields that were updated.
        (default: [])
    :param sanitize_encrypted: If True, encrypted fields will be replaced with
        a constant value (ENCRYPTED_STRING) in the diff. (default: True)
    :param all_values_as_strings: If True, all values will be converted to
        strings after diffing, using Field.value_to_string. (default: False)
    :return: A dictionary with the following
        - added_fields: A dictionary of fields that were added between old and
          new. Importantly, if old and new are the same type of model, this
          should always be empty. An "added field" does not mean that the field
          became non-empty, it means that the field was completely absent from
          the old type of model and is now present in the new type of model. If
          this entry is non-empty, it has the form: {"field_name": value} where
          value is the new value of the field.
        - removed_fields: A dictionary of fields that were removed between old
          and new. Similar to added_fields, if old and new are the same type of
          model, this should always be empty. It has the same structure as
          added_fields except the value is the old value of the field.
        - changed_fields: A dictionary of fields that were changed between old
          and new. The key of each entry is the field name, and the value is a
          tuple of the old value and the new value.
    """

    model_diff = ModelDiff(added_fields={}, removed_fields={}, changed_fields={})

    # Short circuit if both objects are None
    if old is None and new is None:
        return model_diff

    # Fail if we are not dealing with None or Model types
    if (old is not None and not isinstance(old, models.Model)) or (new is not None and not isinstance(new, models.Model)):
        raise TypeError('old and new must be a Model instance or None')

    # If we have to have matching types and both objects are not None and their types don't match then fail
    if require_type_match and (old is not None and new is not None) and type(old) is not type(new):  # noqa: E721
        raise TypeError('old and new must be of the same type')

    # Extract all of the fields and their values into a dict in the format of:
    #  fields = {
    #     'old': { <field>: <value>, [<field>: <value> ...]},
    #     'new': { <field>: <value>, [<field>: <value> ...]},
    #  }
    fields = {}
    for name, obj in (('old', old), ('new', new)):
        fields[name] = {}
        if obj is None:
            continue

        for field in get_all_field_names(obj, concrete_only=True, include_attnames=False):
            field_obj = obj._meta.get_field(field)

            # Skip the field if needed
            if field in exclude_fields:
                continue
            if limit_fields and field not in limit_fields:
                continue
            if not include_m2m and field_obj.many_to_many:
                continue

            if all_values_as_strings:
                if getattr(obj, field) is None:
                    value = None
                else:
                    value = field_obj.value_to_string(obj)
            elif json_safe:
                value = make_json_safe(getattr(obj, field))
            else:
                value = getattr(obj, field)

            fields[name][field] = value

    old_fields_set = set(fields['old'].keys())
    new_fields_set = set(fields['new'].keys())
    old_model = old.__class__ if old else None
    new_model = new.__class__ if new else None

    # Get any removed fields from the old_fields - new_fields
    for field in old_fields_set - new_fields_set:
        model_diff.removed_fields[field] = ENCRYPTED_STRING if is_encrypted_field(old_model, field) else fields['old'][field]

    # Get any new fields from the new_fields - old_fields
    for field in new_fields_set - old_fields_set:
        model_diff.added_fields[field] = ENCRYPTED_STRING if is_encrypted_field(new_model, field) else fields['new'][field]

    # Find any modified fields from the union of the sets
    for field in new_fields_set & old_fields_set:
        if fields['old'][field] != fields['new'][field]:
            model_diff.changed_fields[field] = (
                ENCRYPTED_STRING if is_encrypted_field(old_model, field) else fields['old'][field],
                ENCRYPTED_STRING if is_encrypted_field(new_model, field) else fields['new'][field],
            )

    return model_diff



import logging

from django.utils.translation import gettext_lazy as _
from social_core.backends.azuread import AzureADOAuth2

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.social_auth import SocialAuthMixin, SocialAuthValidateCallbackMixin
from ansible_base.lib.serializers.fields import CharField, URLField

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.azuread')


class AzureADConfiguration(BaseAuthenticatorConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/azuread.html"

    #################################
    # Minimal params
    #################################

    CALLBACK_URL = URLField(
        help_text=_(
            'Provide this URL as the callback URL for your application as part of your registration process. Refer to the documentation for more detail. '
        ),
        ui_field_label=_('Azure AD OAuth2 Callback URL'),
        required=False,
    )

    KEY = CharField(
        help_text=_("The OIDC key (Client ID) from your IDP. Will also be used as the 'audience' for JWT decoding."),
        allow_null=False,
        ui_field_label=_('OIDC Key'),
    )

    SECRET = CharField(
        help_text=_("'The OIDC secret (Client Secret) from your IDP."),
        allow_null=True,
        ui_field_label=_('OIDC Secret'),
    )


class AuthenticatorPlugin(SocialAuthMixin, SocialAuthValidateCallbackMixin, AzureADOAuth2, AbstractAuthenticatorPlugin):
    configuration_class = AzureADConfiguration
    type = "azuread"
    logger = logger
    category = "sso"
    configuration_encrypted_fields = ['SECRET']
